/**
 * \file
 * \copyright (C) 2016-2017 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */


#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <functional>
#include "wx/wx.h"
#include "wx/grid.h"
#include "wx/bookctrl.h"
#include "datafit.h"
#include "misc.h"
#include "customgrids.h"


enum {
	DATACOL_TYPE = 0,
	DATACOL_T,
	DATACOL_T2,
	DATACOL_Y,
	DATACOL_ISRELDEV,
	DATACOL_WEIGHT,
	DATACOL_SERID
};


wxBEGIN_EVENT_TABLE(DataSeriesGrid, wxGrid)
	EVT_GRID_CELL_CHANGING(DataSeriesGrid::OnGridCellChanging)
wxEND_EVENT_TABLE()


const wxString DataSeriesGrid::dataTypeChoices[] = {"Cp", "H"};


DataSeriesGrid::DataSeriesGrid(wxWindow *parent, std::function <void()> cbfFileChanged, wxWindowID id, wxPoint pos, wxSize size) : wxGrid(parent, id, pos, size)
{
	this->cbfFileChanged = cbfFileChanged;
	CreateGrid(0, 7);
	SetColLabelValue(DATACOL_T, wxT("T1 / K"));
	SetColLabelValue(DATACOL_Y, wxT("Y"));
	SetColLabelValue(DATACOL_ISRELDEV, wxT("Rel.Dev."));
	SetColLabelValue(DATACOL_WEIGHT, wxT("\u03C9")); // omega
	SetColLabelValue(DATACOL_T2, wxT("T2 / K"));
	SetColLabelValue(DATACOL_TYPE, wxT("Type"));
	SetColLabelValue(DATACOL_SERID, wxT("Series"));
	SetColFormatBool(DATACOL_ISRELDEV);
 	SetRowLabelSize(40); 
	for (int i = 0; i < 7; i++) {
		SetColSize(i, 80);
	}
}

/**
 * \brief Load existing data to the grid
 */
bool DataSeriesGrid::LoadDataFile(LSQContext &dataFile)
{
	int numOfPoints = dataFile.GetNumOfPoints();
	if (GetNumberRows() > 0) {
		DeleteRows(0, GetNumberRows());
	}

	InsertRows(0, numOfPoints);
	for (int i = 0; i < numOfPoints; i++) {
		Point point = dataFile.GetPoint(i);
		SetCellValue(i, DATACOL_T, wxString::Format("%.14g", point.x));
		SetCellValue(i, DATACOL_Y, wxString::Format("%.14g", point.y));
		SetCellValue(i, DATACOL_ISRELDEV, wxString::Format("%d", point.isRelDev));
		SetCellValue(i, DATACOL_WEIGHT, wxString::Format("%.14g", point.w));
		SetCellValue(i, DATACOL_SERID, point.getSeriesID());
		
		switch (point.dataType) {		
		case DATATYPE_CP:
			SetCellValue(i, DATACOL_TYPE, "Cp");
			SetReadOnly(i, DATACOL_T2, true);
			break;
		case DATATYPE_S:
			SetCellValue(i, DATACOL_TYPE, "S");
			SetReadOnly(i, DATACOL_T2, true);
			break;
		case DATATYPE_H:
			SetCellValue(i, DATACOL_T, wxString::Format("%.14g", point.x1));
			SetCellValue(i, DATACOL_T2, wxString::Format("%.14g", point.x2));
			SetCellValue(i, DATACOL_TYPE, "H");
			SetReadOnly(i, DATACOL_T2, false);
			break;
		}
		SetCellEditor(i, DATACOL_TYPE,
			new wxGridCellChoiceEditor(WXSIZEOF(DataSeriesGrid::dataTypeChoices), DataSeriesGrid::dataTypeChoices, false));
	}
	return true;
}

/**
 * \brief  Reads data from a specially prepared wxGrid class example
 * \param  grid pointer to wxGrid class example
 */
bool DataSeriesGrid::SaveToDataFile(LSQContext &dataFile)
{
	// Check grid size
	if (GetNumberRows() == 0) {
		dataFile.data.clear();
		return true;
	}
	// Convert grid to the data
	int npoints = GetNumberRows();
	dataFile.data.clear();
	for (int i = 0; i < npoints; i++) {
		Point point;
		long int isRelDev;
		GetCellValue(i, DATACOL_T).ToDouble(&point.x);
		GetCellValue(i, DATACOL_Y).ToDouble(&point.y);
		GetCellValue(i, DATACOL_ISRELDEV).ToLong(&isRelDev); point.isRelDev = isRelDev;
		GetCellValue(i, DATACOL_WEIGHT).ToDouble(&point.w);
		point.setSeriesID(GetCellValue(i, DATACOL_SERID));
		wxString dataType = GetCellValue(i, DATACOL_TYPE);
		if (dataType == "Cp") {
			GetCellValue(i, DATACOL_T).ToDouble(&point.x);
			point.dataType = DATATYPE_CP;
		} else if (dataType == "H") {
			GetCellValue(i, DATACOL_T).ToDouble(&point.x1);
			GetCellValue(i, DATACOL_T2).ToDouble(&point.x2);
			point.dataType = DATATYPE_H;
		} else {
			// TODO: ERROR HANDLING
		}

		dataFile.data.push_back(point);
	}
	return true;
}


void DataSeriesGrid::Reset()
{
	if (GetNumberRows() > 0) {
		DeleteRows(0, GetNumberRows());
	}
}

/**
 * \brief Add new experimental point with default values
 */
void DataSeriesGrid::AddPoint()
{
	int ind = GetNumberRows();
	InsertRows(ind, 1);
	SetCellValue(ind, DATACOL_TYPE, "Cp");
	SetCellValue(ind, DATACOL_T, "1");
	SetCellValue(ind, DATACOL_Y, "10");
	SetCellValue(ind, DATACOL_ISRELDEV, "0");
	SetCellValue(ind, DATACOL_WEIGHT, "1");
	SetCellEditor(ind, DATACOL_TYPE,
		new wxGridCellChoiceEditor(WXSIZEOF(DataSeriesGrid::dataTypeChoices), DataSeriesGrid::dataTypeChoices, false));
	SetReadOnly(ind, DATACOL_T2, true);
}

/**
 * \brief Delete existing point(s) (selected rows)
 */
void DataSeriesGrid::DeleteSelectedPoints()
{
	wxArrayInt rows = GetSelectedRows();
	for (int i = rows.GetCount() - 1; i >= 0; i--) {
		DeleteRows(rows[i], 1);
	}
}

/**
 * \brief  Set all deviations in the table to relative
 */
void DataSeriesGrid::SetRelDev()
{
	for (int i = 0; i < GetNumberRows(); i++) {
		SetCellValue(i, DATACOL_ISRELDEV, "1");
	}
}

/**
 * \brief  Set all deviations in the table to absolute
 */
void DataSeriesGrid::SetAbsDev()
{
	for (int i = 0; i < GetNumberRows(); i++) {
		SetCellValue(i, DATACOL_ISRELDEV, "0");
	}
}

wxPoint DataSeriesGrid::GetPosition()
{
	return wxGrid::GetPosition();
}

void DataSeriesGrid::SetSize(int width, int height)
{
	wxGrid::SetSize(width, height);
}

void DataSeriesGrid::ResizeCols()
{
	ResizeWxGridCols(this);
}

/**
 * \brief Controls user input and prevents changes if it is incorrect
 */
void DataSeriesGrid::OnGridCellChanging(wxGridEvent& event)
{
	wxString strNotNumeric = wxString::FromUTF8("Not numeric parameter");
	int col = event.GetCol();
	double t;
	if (col == DATACOL_T || col == DATACOL_T2 || col == DATACOL_Y) {
		if (!event.GetString().ToDouble(&t) || t <= 0.0) {
			wxMessageBox(strNotNumeric, "CpFit", wxICON_ERROR);
			event.Veto();
		}
	} else if (col == DATACOL_WEIGHT) {
		if (!event.GetString().ToDouble(&t) || t < 0.0) {
			wxMessageBox(strNotNumeric, "CpFit", wxICON_ERROR);
			event.Veto();
		}
	} else if (col == DATACOL_TYPE) { // data type row
		wxString dataType = event.GetString();
		int row = event.GetRow();
		if (dataType == "Cp") {
			SetReadOnly(row, DATACOL_T2); // Disable T2 column
		} else if (dataType == "H") {
			SetReadOnly(row, DATACOL_T2, false); // Enable T2 column
		} else {
			event.Veto();
		}
	}
	// We don't check DATACOL_ISRELDEV because if is boolean
	// We don't check DATACOL_SERID because it contains series ID in free form
	cbfFileChanged();
}


///////////////////////////////////////////////
/**
 * \brief Controls IDs
 */
enum {
	ID_BTN_CALC = 100,
	ID_CMB_PARNUM,
	ID_BTN_RESETPAR,
	ID_BTN_ADDPOINT,
	ID_BTN_DELPOINT,
	ID_BTN_SETREL,
	ID_BTN_SETABS,
	ID_GR_DATA,
};


wxBEGIN_EVENT_TABLE(DataSeriesEditor, wxPanel)
	EVT_SIZE(DataSeriesEditor::OnSize)
	EVT_BUTTON(ID_BTN_ADDPOINT, DataSeriesEditor::OnAddPoint)
	EVT_BUTTON(ID_BTN_DELPOINT, DataSeriesEditor::OnDeleteSelectedPoints)
	EVT_BUTTON(ID_BTN_SETREL, DataSeriesEditor::OnSetRelDev)
	EVT_BUTTON(ID_BTN_SETABS, DataSeriesEditor::OnSetAbsDev)
wxEND_EVENT_TABLE()

DataSeriesEditor::DataSeriesEditor(wxWindow *parent, std::function <void()> cbfFileChanged, wxPoint pos, wxSize size) : wxPanel(parent, wxID_ANY, pos, size)
{
	this->cbfFileChanged = cbfFileChanged;
	grData = new DataSeriesGrid(this, this->cbfFileChanged, ID_GR_DATA, wxPoint(0, 0), wxSize(380, 230));
	wxSize datBtnSize(100, 30);
	wxPoint datBtnPos(0, 260);
	btnAddPoint = new wxButton(this, ID_BTN_ADDPOINT, "Add point",    datBtnPos, datBtnSize);
	datBtnPos.x += datBtnSize.x + 10;
	btnDelPoint = new wxButton(this, ID_BTN_DELPOINT, "Delete point", datBtnPos, datBtnSize);
	datBtnPos.x += datBtnSize.x + 10;
	btnSetAbs   = new wxButton(this, ID_BTN_SETABS, "Set abs.dev.", datBtnPos, datBtnSize);
	datBtnPos.x += datBtnSize.x + 10;
	btnSetRel   = new wxButton(this, ID_BTN_SETREL, "Set rel.dev.", datBtnPos, datBtnSize);
	datBtnPos.x += datBtnSize.x + 10;
}

void DataSeriesEditor::OnSize(wxSizeEvent &event)
{
	// Set coordinates of elements inside boxData (experimental data)
	wxSize s = GetSize();
	btnAddPoint->Move(btnAddPoint->GetPosition().x, s.y - 30);
	btnDelPoint->Move(btnDelPoint->GetPosition().x, s.y - 30);
	btnSetAbs->Move(btnSetAbs->GetPosition().x, s.y - 30);
	btnSetRel->Move(btnSetRel->GetPosition().x, s.y - 30);
	grData->SetSize(s.x, btnSetAbs->GetPosition().y - grData->GetPosition().y - 20);
	// Resize columns inside grPar and gr Data tables
	grData->ResizeCols();
}

bool DataSeriesEditor::LoadDataFile(LSQContext &dataFile)
{
	return grData->LoadDataFile(dataFile);
}

bool DataSeriesEditor::SaveToDataFile(LSQContext &dataFile)
{
	return grData->SaveToDataFile(dataFile);
}

void DataSeriesEditor::Reset()
{
	grData->Reset();
}

/**
 * \brief Add new experimental point
 */
void DataSeriesEditor::OnAddPoint(wxCommandEvent &event)
{
	grData->AddPoint();
	cbfFileChanged();
}

/**
 * \brief Delete existing point(s) (selected rows)
 */
void DataSeriesEditor::OnDeleteSelectedPoints(wxCommandEvent &event)
{
	grData->DeleteSelectedPoints();
	cbfFileChanged();
}

/**
 * \brief  Set all deviations in the table to relative
 */
void DataSeriesEditor::OnSetAbsDev(wxCommandEvent &event)
{
	grData->SetAbsDev();
	cbfFileChanged();
}

/**
 * \brief  Set all deviations in the table to relative
 */
void DataSeriesEditor::OnSetRelDev(wxCommandEvent &event)
{
	grData->SetRelDev();
	cbfFileChanged();
}

void DataSeriesEditor::ResizeCols()
{
	grData->ResizeCols();
}


//////////////////////////////////////////////
enum {
	ID_BTN_ADDTERM = 100,
	ID_BTN_DELTERM,
	ID_CMB_TERMSLIST,
	ID_CMB_TERMTYPE,
	ID_GR_FIXPARAM,
	ID_GR_LSQPARAM,
	ID_TXT_TMIN,
	ID_TXT_TMAX
};

wxBEGIN_EVENT_TABLE(ExtraTermsEditor, wxPanel)
	EVT_SIZE(ExtraTermsEditor::OnSize)
	EVT_BUTTON(ID_BTN_ADDTERM, ExtraTermsEditor::OnAddTerm)
	EVT_BUTTON(ID_BTN_DELTERM, ExtraTermsEditor::OnDeleteTerm)
	EVT_COMBOBOX(ID_CMB_TERMSLIST, ExtraTermsEditor::OnCmbTermsList)
	EVT_COMBOBOX(ID_CMB_TERMTYPE, ExtraTermsEditor::OnCmbTermType)
	EVT_GRID_CELL_CHANGING(ExtraTermsEditor::OnGridCellChanging)
	EVT_TEXT(ID_TXT_TMIN, ExtraTermsEditor::OnTminChange)
	EVT_TEXT(ID_TXT_TMAX, ExtraTermsEditor::OnTmaxChange)
wxEND_EVENT_TABLE()

ExtraTermsEditor::ExtraTermsEditor(wxWindow *parent, std::function <void()> cbfFileChanged, wxPoint pos, wxSize size)
	: wxPanel(parent, wxID_ANY, pos, size)	
{
	wxString strTermTypes[] = {"Default", "Two peaks"};
	wxPoint posLbl(10, 10);
	wxPoint posTxt(90, 10);
	this->cbfFileChanged = cbfFileChanged;
	// Terms selection menu
	wxStaticText *lblTermsList = new wxStaticText(this, wxID_ANY, "Terms:", posLbl, wxSize(80, 30));
	cbTermsList = new wxComboBox(this, ID_CMB_TERMSLIST, "", posTxt, wxSize(100, 30));
	btnAddTerm = new wxButton(this, ID_BTN_ADDTERM, "Add", wxPoint(200, posLbl.y), wxSize(70, 30));
	btnDelTerm = new wxButton(this, ID_BTN_DELTERM, "Delete", wxPoint(270, posLbl.y), wxSize(70, 30));
	posLbl.y += 35; posTxt.y += 35;
	// Term editor
	// a) type field
	wxStaticText *lblTermType = new wxStaticText(this, wxID_ANY, "Term type:", posLbl, wxSize(80, 30));
	cbTermType = new wxComboBox(this, ID_CMB_TERMTYPE, strTermTypes[0], posTxt, wxSize(100, 30), 2, strTermTypes, wxCB_READONLY);
	// b) formula field
 	bmpFormula = new wxStaticBitmap(this, wxID_ANY, wxBITMAP(term_default), wxPoint(posTxt.x + 110, posLbl.y), wxSize(250, 30));
	posLbl.y += 35; posTxt.y += 35;
	// c) Tmin field
	wxStaticText *lblTmin = new wxStaticText(this, wxID_ANY, "Tmin, K:", posLbl, wxSize(80, 30));
	txtTmin = new wxTextCtrl(this, ID_TXT_TMIN, "", posTxt, wxSize(100, 30));
	// c) Tmax field
	wxStaticText *lblTmax = new wxStaticText(this, wxID_ANY, "Tmax, K:", wxPoint(posLbl.x + 200, posLbl.y), wxSize(80, 30));
	txtTmax = new wxTextCtrl(this, ID_TXT_TMAX, "", wxPoint(posLbl.x + 300, posLbl.y), wxSize(100, 30));
	posLbl.y += 35; posTxt.y += 35;
	// d) Constants grid
	boxLSQParam = new wxStaticBox(this, wxID_ANY, "LSQ parameters", wxPoint(posLbl.x + 170, posLbl.y), wxSize(150, 200));
	boxFixParam = new wxStaticBox(this, wxID_ANY, "Fixed parameters", wxPoint(posLbl.x, posLbl.y), wxSize(150, 200));
	grFixParam = new wxGrid(this, ID_GR_FIXPARAM, wxPoint(20, 30), wxSize(150, 150));
	grFixParam->CreateGrid(0, 2);
	grFixParam->SetRowLabelSize(40);
	grFixParam->SetColLabelValue(0, "Name");
	grFixParam->SetColLabelValue(1, "Value");
	// e) LSQ parameters grid
	grLSQParam = new wxGrid(this, ID_GR_LSQPARAM, wxPoint(20, 30), wxSize(150, 150));
	grLSQParam->CreateGrid(0, 2);
	grLSQParam->SetRowLabelSize(40);
	grLSQParam->SetColLabelValue(0, "Name");
	grLSQParam->SetColLabelValue(1, "Value");
	// Some default settings
	currentTermIndex = -1; // Nothing selected
	ClearEditor();
	DisableEditor();
}

/**
 * \brief Removes all extra terms from the internal buffer
 * and clears the form
 */
void ExtraTermsEditor::Reset()
{
	termsBuf.clear();
	LoadTerms(termsBuf);
}

/**
 * \brief Clear the form (Tmin and Tmax values, table, formula bitmap etc.)
 */
void ExtraTermsEditor::ClearEditor()
{
	std::vector <double> emptyVector;
	VecToTblRow(grFixParam, emptyVector);
	VecToTblRow(grLSQParam, emptyVector);

	wxBitmap emptyBitmap;
	emptyBitmap.Create(0, 0);
	bmpFormula->SetBitmap(emptyBitmap);

	txtTmin->ChangeValue("");
	txtTmax->ChangeValue("");
}

/**
 * \brief Enable/disable form for editing the extra term
 */
void ExtraTermsEditor::EnableEditor(bool enable)
{	
	txtTmin->Enable(enable);
	txtTmax->Enable(enable);
	cbTermType->Enable(enable);
	grFixParam->Enable(enable);
	grLSQParam->Enable(enable);
	btnDelTerm->Enable(enable);
}

void ExtraTermsEditor::DisableEditor()
{
	EnableEditor(false);
}

/**
 * \brief Controls grLSQParam and grFixParam wxGrid objects
 *
 * Controls data type (only doubles are accepted) and values ranges for grLSQParam
 */
void ExtraTermsEditor::OnGridCellChanging(wxGridEvent& event)
{
	// Data type control
	double t;
	bool changed = true;
	if (!event.GetString().ToDouble(&t)) {
		wxMessageBox("Not numeric parameter", "CpFit", wxICON_ERROR);
		changed = false;
	}
	// Grid-specific things
	if (event.GetId() == ID_GR_LSQPARAM) {
		int row = event.GetRow();
		if (row < lsqparam_lb.size() && (t < lsqparam_lb[row] || t > lsqparam_ub[row]) ) {
			wxMessageBox(
				wxString::Format("Parameter boundaries are [%g;%g]", lsqparam_lb[row], lsqparam_ub[row]),
				"CpFit",
				wxICON_ERROR);
			changed = false;
		}
	}
	// Change control
	if (changed) { // New value is correct: set "file is changed" flag
		cbfFileChanged();
	} else { // New value is incorrect: cancel editing
		event.Veto();
	}
}

/**
 * \brief Used by ExtraTermsEditor::OnTminChange and ExtraTermsEditor::OnTmaxChange
 * for control of T value (must be a real number between 0 and 20000 K)
 */
double ExtraTermsEditor::OnTChange(wxString str)
{
	double t = -1;
	if (!str.ToDouble(&t)) {
		wxMessageBox("Not numeric parameter", "CpFit", wxICON_ERROR);
		t = -1;
	} else if (t < 0 || t > 20000) {
		wxMessageBox("Invalid T value", "CpFit", wxICON_ERROR);
		t = -1;
	};
	return t;
}

/**
 * \brief Controls Tmin value (in GUI) change
 */
void ExtraTermsEditor::OnTminChange(wxCommandEvent &event)
{	
	double t = OnTChange(event.GetString());
	if (t == -1) {
		txtTmin->ChangeValue(wxString::Format("%g", termsBuf[currentTermIndex].Tmin));
	} else {
		cbfFileChanged();
	}
}

/**
 * \brief Controls Tmax value (in GUI) change
 */
void ExtraTermsEditor::OnTmaxChange(wxCommandEvent &event)
{
	double t = OnTChange(event.GetString());
	if (t == -1) {
		txtTmax->ChangeValue(wxString::Format("%g", termsBuf[currentTermIndex].Tmax));
	} else {
		cbfFileChanged();
	}
}

/**
 * \brief Converts double vector to wxGrid row
 */
void ExtraTermsEditor::VecToTblRow(wxGrid *gr, std::vector <double> &p)
{
	int nparams = p.size();
	int nrows = gr->GetNumberRows();
	if (nrows > 0) {
		gr->DeleteRows(0, gr->GetNumberRows());
	}
	gr->InsertRows(0, nparams);	
	for (int i = 0; i < p.size(); i++) {
		gr->SetCellValue(i, 1, wxString::Format("%g", p[i]));
	}
}

/**
 * \brief Converts wxGrid row to double vector
 */
std::vector <double> ExtraTermsEditor::TblRowToVec(wxGrid *gr)
{
	std::vector <double> vec;
	int nrows = gr->GetNumberRows();
	for (int i = 0; i < nrows; i++) {
		double val;
		gr->GetCellValue(i, 1).ToDouble(&val);
		vec.push_back(val);
	}
	return vec;
}


/**
 * \brief Loads data from ExtraTerm class example into GUI
 */
bool ExtraTermsEditor::LoadTerm(ExtraTerm &term)
{
	// Load bitmap with formula for selected term
	if (term.GetType() == EXTRATERM_DEFAULT) {	
		cbTermType->SetSelection(0);
		bmpFormula->SetBitmap(wxBITMAP(term_default));		
	} else if (term.GetType() == EXTRATERM_TWOPEAKS) {
		cbTermType->SetSelection(1);
		bmpFormula->SetBitmap(wxBITMAP(term_twopeaks));
	}
	Refresh();
	// Load term parameters
	txtTmin->SetValue(wxString::Format("%g", term.Tmin));
	txtTmax->SetValue(wxString::Format("%g", term.Tmax));
	std::vector <double> fixparam = term.GetFixParam();
	std::vector <double> lsqparam = term.GetLSQParam();
	VecToTblRow(grFixParam, fixparam);
	VecToTblRow(grLSQParam, lsqparam);
	lsqparam_lb = term.GetLSQParamLB();
	lsqparam_ub = term.GetLSQParamUB();
	// Load names of parameters
	std::vector <std::string> fixparam_names = term.GetFixParamNames();
	std::vector <std::string> lsqparam_names = term.GetLSQParamNames();
	for (int i = 0; i < fixparam_names.size(); i++) {
		grFixParam->SetCellValue(i, 0, fixparam_names[i]);
	}
	for (int i = 0; i < lsqparam_names.size(); i++) {
		grLSQParam->SetCellValue(i, 0, lsqparam_names[i]);
	}
	return true;
}

/**
 * \brief Save data from GUI to ExtraTerm class examle
 */
bool ExtraTermsEditor::SaveTerm(ExtraTerm &term)
{
	ExtraTerm ans(GetGUITermType());
	if (!txtTmin->GetValue().ToDouble(&ans.Tmin)) {
		return false;
	}
	if (!txtTmax->GetValue().ToDouble(&ans.Tmax)) {
		return false;
	}
	std::vector <double> lsqparam = TblRowToVec(grLSQParam);
	std::vector <double> fixparam = TblRowToVec(grFixParam);
	ans.SetLSQParam(lsqparam);
	ans.SetFixParam(fixparam);
	
	term = ans;	
	return true;
}

ExtraTermType ExtraTermsEditor::GetGUITermType()
{
	ExtraTermType termType = EXTRATERM_DEFAULT;
	int ind = cbTermType->GetSelection();	
	if (ind < 0) {
		ind = 0;
	}
	if (ind == EXTRATERM_DEFAULT) {
		termType = EXTRATERM_DEFAULT;
	} else if (ind == EXTRATERM_TWOPEAKS) {
		termType = EXTRATERM_TWOPEAKS;
	}
	return termType;
}

/**
 * \brief Loads ExtraTerm class vector into ExtraTermsEditor internal variables
 * and shows the first term at the page
 */
bool ExtraTermsEditor::LoadTerms(std::vector <ExtraTerm> &terms)
{
	termsBuf = terms;
	// a) Initialize combo box
	cbTermsList->Clear();
	for (int i = 0; i < termsBuf.size(); i++) {
		cbTermsList->Append(wxString::Format("%d", i + 1));
	}
	if (termsBuf.size() > 0) {
		cbTermsList->SetSelection(0);
		LoadTerm(termsBuf[0]);
		EnableEditor();
		currentTermIndex = 0;
	} else {
		ClearEditor();
		DisableEditor();
	}
	return false;
}

bool ExtraTermsEditor::SaveTerms(std::vector <ExtraTerm> &terms)
{
	terms = termsBuf;
	return false;
}

bool ExtraTermsEditor::LoadDataFile(LSQContext &dataFile)
{
	return LoadTerms(dataFile.extraTerms);
}

bool ExtraTermsEditor::SaveToDataFile(LSQContext &dataFile)
{
	if (currentTermIndex != -1) {
		// Save current GUI content
		SaveTerm(termsBuf[currentTermIndex]);
	}
	return SaveTerms(dataFile.extraTerms);
}


/**
 * \brief Resizes boxes and grids with constants and LSQ parameters
 * initial values
 */
void ExtraTermsEditor::OnSize(wxSizeEvent &event)
{
	wxSize nbkSize = event.GetSize();
	wxPoint fixPos = boxFixParam->GetPosition();
	wxSize gridSize(nbkSize.x / 2 - 15, nbkSize.y - fixPos.y - 15);
	// wxStaticBox sizes and positions	
	boxFixParam->Move(10, fixPos.y);
	boxLSQParam->Move(nbkSize.x / 2 + 5, fixPos.y);	
	boxFixParam->SetSize(gridSize);
	boxLSQParam->SetSize(gridSize);
	// wxGrid sizes and positions	
	grFixParam->SetSize(gridSize.x - 50, gridSize.y - 50);
	grLSQParam->SetSize(gridSize.x - 50, gridSize.y - 50);	
	grFixParam->Move(20, fixPos.y + 30);
	grLSQParam->Move(boxLSQParam->GetPosition().x + 20, fixPos.y + 30);
	// wxGrid columns resize	
	ResizeWxGridCols(grFixParam);
	ResizeWxGridCols(grLSQParam);
}

void ExtraTermsEditor::OnAddTerm(wxCommandEvent &event)
{
	termsBuf.push_back(ExtraTerm());
	wxString termName = wxString::Format("%d", termsBuf.size());
	cbTermsList->Append(termName);
	cbTermsList->SetValue(termName);
	if (currentTermIndex != -1) {
		// Save current GUI content
		SaveTerm(termsBuf[currentTermIndex]);
	}
	currentTermIndex = termsBuf.size() - 1;
	LoadTerm(termsBuf[currentTermIndex]);
	cbTermsList->SetSelection(termsBuf.size() - 1);
	EnableEditor();
	cbfFileChanged();
}

void ExtraTermsEditor::OnDeleteTerm(wxCommandEvent &event)
{
	int ind = cbTermsList->GetSelection();
	if (termsBuf.size() > 0 && ind >= 0) {
		wxString strMsg = wxString::Format("Do you want to delete extra term %d?", ind + 1);
		int ans = wxMessageBox(strMsg, "CpFit", wxYES_NO | wxICON_EXCLAMATION);
		if (ans == wxYES) {
			termsBuf.erase(termsBuf.begin() + ind);
			LoadTerms(termsBuf);
			if (cbTermsList->GetCount() > 0) {
				cbTermsList->SetSelection(0);
			}
			cbfFileChanged();
		}
	}
}

/**
 * \params Shows selected term parameters in the GUI
 */
void ExtraTermsEditor::OnCmbTermsList(wxCommandEvent &event)
{
	int ind = cbTermsList->GetSelection();
	if (currentTermIndex != -1) {
		// Save GUI content
		SaveTerm(termsBuf[currentTermIndex]);
	}
	LoadTerm(termsBuf[ind]);
	currentTermIndex = ind;
}

/**
 * \params Processes selection of another term type in the GUI
 * (e.g. parameters sizes etc.)
 */
void ExtraTermsEditor::OnCmbTermType(wxCommandEvent &event)
{
	// Update term type
	if (currentTermIndex != -1) {
		// Save GUI content
		termsBuf[currentTermIndex].SetTermType(GetGUITermType());
		SaveTerm(termsBuf[currentTermIndex]);
	}
	// Load data for selected term
	LoadTerm(termsBuf[currentTermIndex]);
	// File changed
	cbfFileChanged();
}



/////////////////////////////////////////////////////////
DataFileEditor::DataFileEditor(wxWindow *parent, std::function <void()> cbfFileChanged, wxPoint pos, wxSize size)
	: wxNotebook(parent, wxID_ANY, pos, size)
{
	this->cbfFileChanged = cbfFileChanged;
	ctlData = new DataSeriesEditor(this, this->cbfFileChanged, wxPoint(20, 30), wxSize(380, 230));
	ctlExtraTerms = new ExtraTermsEditor(this, this->cbfFileChanged, wxPoint(20, 30), wxSize(380, 230));
	AddPage(ctlData, "Data points");
	AddPage(ctlExtraTerms, "Extra terms");
}

bool DataFileEditor::LoadDataFile(LSQContext &dataFile)
{
	if (!ctlData->LoadDataFile(dataFile)) {
		return false;
	}
	if (!ctlExtraTerms->LoadDataFile(dataFile)) {
		return false;
	}
	return true;
}

bool DataFileEditor::SaveToDataFile(LSQContext &dataFile)
{
	if (!ctlData->SaveToDataFile(dataFile)) {
		return false;
	}
	if (!ctlExtraTerms->SaveToDataFile(dataFile)) {
		return false;
	}
	return true;
}

void DataFileEditor::Reset()
{
	ctlData->Reset();
	ctlExtraTerms->Reset();
}
