/**
 * \file
 * \copyright (C) 2016-2017 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <time.h>
#include <string>
#include <vector>
#include <algorithm>
#include <functional>
#include "datafit.h"
#include "dfunc3.h"
#include "levmar.h"


// Import internal function from LEVMAR library that calculates
// covariance matrix
extern "C" int dlevmar_covar(double *JtJ, double *C, double sumsq, int m, int n);


/**
 * \brief Cumulative t-distribution function
 * \param x Cumulative distribution function argument
 * \param f Degrees of freedom
 *
 * The function implementation is based on the next article:
 * Hubert J Chen. On approximations to the inverse student's t-distribution function //
 * // Journal of Statistical Computation and Simulation, 1978, 7:3-4, 167-180
 * DOI: http://dx.doi.org/10.1080/00949657808810225
 */
double tdistr_cdf(double x, int f)
{
	// Check input arguments
	if (f <= 0) { // Invalid value
		return -1;
	}
	// Intermediate variables
	double t = atan(x / sqrt(f)); // theta
	double cos_t = cos(t); // cos(theta)
	double F = 0.5;
	// Calculate series
	if (f == 1) {
		F += t / M_PI;
	} else if (f % 2 == 1) { // Odd
		double cossum = cos_t, coeff = 1.0;
		for (int i = 3; i <= f - 2; i += 2) {
			coeff *= (double) (i - 1) / i;
			cossum += coeff*pow(cos_t, i);
		}
		F += (t + sin(t)*cossum) / M_PI;		
	} else if (f % 2 == 0) { // Even
		double cossum = 1.0, coeff = 1.0;
		for (int i = 2; i <= f - 2; i += 2) {
			coeff *= (double) (i - 1) / i;
			cossum += coeff*pow(cos_t, i);
		}
		F += sin(t) / 2 * cossum;
	}
	// Return the function value
	return F;
}

/**
 * \brief Signum function
 */
double sign(double x)
{
	if (x == 0) {
		return 0;
	} else if (x > 0) {
		return 1;
	} else {
		return -1;
	}
}

/**
 * \brief Inverse t-distribution function
 * \param p Probability (0 - 1)
 * \param f Degrees of freedom
 *
 * Uses bisection method and tdistr_cdf function
 */
double tdistr_inv(double p, double f)
{
	// Input arguments control
	if (p < 0 | p > 1) {
		return -1;
	}
	if (p < 1e-12) {
		return -INFINITY;
	}
	if (p > 1 - 1e-12) {
		return INFINITY;
	}
	// Initialize variables	
	double tA = -50.0, tB = 50.0, tC;
	double FA, FB;
	// Find boundaries for bisection method
	do {	
		tA *= 2;
		tB *= 2;
		FA = tdistr_cdf(tA, f) - p;
		FB = tdistr_cdf(tB, f) - p;
		if (FA == 0) {
			return tA;
		}
		if (FB == 0) {
			return tB;
		}
	} while (sign(FA) * sign(FB) > 0);
	// Bisection method iterations
	int iter = 0;
	while (fabs(FB - FA) > DBL_EPSILON && iter < 10000) {
		iter++;
		tC = tA + 0.5*(tB - tA);
		double FC = tdistr_cdf(tC, f) - p;

		if (FC == 0) {
			return tC;
		}

		if (sign(FA)*sign(FC) < 0) {
			tB = tC; 
			FB = FC;
		}
		else if (sign(FC)*sign(FB) < 0) {
			tA = tC;
			FA = FC;
		}
	}
	return tC;
}

/**
 * \brief exp(-x)*Ei(x) function
 */
double eminusx_ei(double x)
{
	double val = 1, fact = 1, d = 1;
	if (x <= -2) { // Continued fraction
		for (int i = 100; i >= 1; i--) {
			val = (-x) + i / (1 + i / val);
		}
		val = -1.0 / val;
	} else if (fabs(x) < fabs(log(DBL_EPSILON))) { // Taylor series
		val = exp(-x) * (EULER_GAMMA + log(fabs(x)));
		for (int i = 1; i < 500 && fabs(d)/fabs(val) > DBL_EPSILON; i++) {
			fact *= i;
			d = exp(-x) * pow(x, i) / (i * fact);
			val += d;
		}
	} else { // Asymptotic formula
		double d_prev = 1e100;
		for (int i = 1; i < 500 && fabs(d)/fabs(val) > DBL_EPSILON && fabs(d) < fabs(d_prev); i++) {
			fact *= i;
			d_prev = d;
			d = fact / pow(x, i);
			val += d;
		}
		val /= x;
	}
	return val;
}




/**
 * \brief Constructor creates a point with invalid x and y values,
 * absolute deviation, Cp data type and w = 1
 */
Point::Point()
{
	x = -1.0;
	x1 = -1.0;
	x2 = -1.0;
	y = -1.0;
	isRelDev = 0;
	w = 1.0;
	dataType = DATATYPE_CP;
	seriesID[0] = 0; // ASCIIZ-string
}

void Point::setSeriesID(const char *id)
{
	int i;
	for (i = 0; i < 15 && id[i] != 0; i++) {
		seriesID[i] = id[i];
	}
	seriesID[i] = 0; // End of line
}

std::string Point::getSeriesID()
{
	return std::string(seriesID);
}



// ===== TDPEFunc constructors =====

/**
 * \brief Class example with no model parameters
 */
TDPEFunc::TDPEFunc()
{
}

/**
 * \brief Class example with user-defined model parameters
 */
TDPEFunc::TDPEFunc(double *param, int nparams)
{
	this->param = std::vector <double> (param, param + nparams);
}

/**
 * \brief Returns current parameters set
 */
std::vector <double> TDPEFunc::GetParam()
{
	return param;
}

/**
 * \brief Returns parameters set in the form of two vectors (alpha and theta)
 */
std::pair < std::vector <double>, std::vector<double> > TDPEFunc::GetAlphaTheta()
{
	std::vector <double> alpha;
	std::vector <double> theta;
	int nparams = param.size();

	for (int i = 0; i < nparams; i += 2) {
		alpha.push_back(param[i]);
		theta.push_back(param[i+1]);
	}
	return std::make_pair(alpha, theta);
}


int TDPEFunc::GetNumOfParams()
{
	return param.size();
}

void TDPEFunc::AddTerm(double alpha, double theta)
{
	param.push_back(alpha);
	param.push_back(theta);
}


// ===== TDPEFunc methods =====
/**
 * \brief Cp(T) function implementation (internal parameters used)
 * \param T Temperature, K
 */
std::vector <double> TDPEFunc::CpFunc(std::vector <double> T)
{
	int npoints = T.size();
	std::vector <double> Cp(npoints);
	for (int i = 0; i < npoints; i++) {
		Cp[i] = CpFunc(T[i]);
	}
	return Cp;
}

/**
 * \brief S(T)-S(0) function implementation (internal parameters used)
 * \param T Temperature, K
 */
std::vector <double> TDPEFunc::SFunc(std::vector <double> T)
{
	int npoints = T.size();
	std::vector <double> S(npoints);
	for (int i = 0; i < npoints; i++) {
		S[i] = SFunc(T[i]);
	}
	return S;
}

/**
 * \brief H(T)-H(0) function implementation (internal parameters used)
 * \param T Temperature, K
 */
std::vector <double> TDPEFunc::HFunc(std::vector <double> T)
{
	int npoints = T.size();
	std::vector <double> H(npoints);
	for (int i = 0; i < npoints; i++) {
		H[i] = HFunc(T[i]);
	}
	return H;
}

/**
 * \brief H(T2)-H(T1) function implementation (internal parameteres used)
 * \param T1 Lower temperature, K
 * \param T2 Higher temperatre, K
 */
std::vector <double> TDPEFunc::DHFunc(std::vector <double> T1, std::vector <double> T2)
{
	int npoints = T1.size();
	std::vector <double> DH(npoints);
	if (T1.size() != T2.size()) {
		return DH;
	}
	for (int i = 0; i < npoints; i++) {
		DH[i] = DHFunc(T1[i], T2[i]);
	}
	return DH;
}

/**
 * \brief Cp(T) function implementation (internal parameters used)
 * \param T Temperature, K
 */
double TDPEFunc::CpFunc(double T)
{
	double Cp = 0.0;
	int nparams = param.size();
	// a) E.-P. functions part
	for (int j = 0; j < nparams; ) {
		double alpha = param[j++];
		double theta = param[j++];
		double x = theta / T;
		double ex = exp(x);

		if (x < 500) {
			Cp += 3*alpha*ex*x*x / pow(ex - 1, 2);
		} else {
			Cp += 3*alpha*x*x / ex;
		}
	}
	Cp = Cp * CONST_R;
	// b) Extra terms part
	int nextraterms = extraTerms.size();
	for (int j = 0; j < nextraterms; j++) {
		Cp += extraTerms[j].CpFuncEval(T);
	}
	return Cp;
}

/**
 * \brief (&part;Cp/&part;&beta;) derivatives calculation
 * (required for Jacobian calculation)
 * \param T temperature, K
 * \param J pointer to the Jacobian buffer (must have at least nparams
 * free elements)
 *
 * This function is actively used by LSQContext::GetJVector to calculate
 * Jacobian. &beta; is generalized notation for model parameters, i.e.
 *
 * &beta;[0] = &alpha;[0], &beta;[1] = &theta;[0], &beta;[2] = &alpha;[1], &beta;[3] = &theta;[1] etc.
 *
 * Order of derivatives in memory correspond this rule
 */
void TDPEFunc::dCpdBFunc(double T, double *J)
{
	int nparams = param.size();
	for (int j = 0; j < nparams; ) {
		double alpha = param[j++];
		double theta = param[j++];
		double x = theta / T;
		double ex = exp(x);

		if (x < 500) {
			*J++ = 3*ex*x*x / pow(ex - 1, 2) * CONST_R;
			*J++ = 3*alpha / (T * pow(ex - 1, 2)) * x * ex * ( 
				(2 + x) - 2*x*ex / (ex - 1)
			) * CONST_R;
		} else {
			*J++ = 3*x*x / (ex) * CONST_R;
			*J++ = 3*alpha*x / (T * ex) * (2 - x) * CONST_R;
		}
	}
}


/**
 * \brief (&part;H/&part;&beta;) derivatives calculation
 * (required for Jacobian calculation)
 * \param T temperature, K
 * \param J pointer to the Jacobian buffer (must have at least nparams
 * free elements)
 *
 * This function is actively used by LSQContext::GetJVector to calculate
 * Jacobian. See TDPEFunc::param for model parameters order. 
 * Order of derivatives in memory corresponds this rule
 */
void TDPEFunc::dHdBFunc(double T, double *J)
{
	for (int j = 0; j < param.size(); j += 2) {
		double alpha = param[j];
		double theta = param[j+1];
		double x = theta / T;
		double ex = exp(x);

		if (x < 500) {
			*J++ = 3 * x / (exp(x) - 1) * T * CONST_R; 
			*J++ = 3 * alpha * (ex - 1 - x*ex) / pow(ex - 1, 2) * CONST_R;
		} else {
			*J++ = 3 * x / (exp(x) - 1) * T * CONST_R; 
			*J++ = 3 * alpha * (1 - x) / ex * CONST_R;
		}
	}
}

/**
 * \brief (&part;S/&part;&beta;) derivatives calculation
 * (required for Jacobian calculation)
 * \param T temperature, K
 * \param J pointer to the Jacobian buffer (must have at least nparams
 * free elements)
 *
 * This function is actively used by LSQContext::GetJVector to calculate
 * Jacobian. See TDPEFunc::param for model parameters order. 
 * Order of derivatives in memory corresponds this rule.
 */
void TDPEFunc::dSdBFunc(double T, double *J)
{
	for (int j = 0; j < param.size(); j += 2) {
		double alpha = param[j];
		double theta = param[j+1];
		double x = theta / T;
		double ex = exp(x);

		if (x < 500) {		
			*J++ = 3*(x / (ex - 1) - log(1 - exp(-x))) * CONST_R;
			*J++ = -3*alpha*theta * ex / pow(T * (ex - 1), 2) * CONST_R;
		} else {
			*J++ = 3*(x / (ex - 1) - log(1 - exp(-x))) * CONST_R;
			*J++ = -3*alpha*theta / ex * CONST_R;
		}
	}
}

/**
 * \brief S(T)-S(0) function implementation (internal parameters used)
 * \param T Temperature, K
 */
double TDPEFunc::SFunc(double T)
{
	double S = 0.0;
	// a) E.-P. part
	for (int j = 0; j < param.size(); j += 2) {
		double alpha = param[j];
		double theta = param[j+1];
		double x = theta / T;
		double ex = exp(x);		
		// TODO: FOR X > 500
		S += 3*alpha*(x / (ex - 1) - log(1 - exp(-x)));
	}
	S *= CONST_R;
	// b) Extra terms part
	int nextraterms = extraTerms.size();
	for (int j = 0; j < nextraterms; j++) {
		S += extraTerms[j].SFuncEval(T);
	}
	return S;
}

/**
 * \brief H(T)-H(0) function implementation (internal parameters used)
 * \param T Temperature, K
 */
double TDPEFunc::HFunc(double T)
{
	double H = 0.0;
	// a) E.-P. part
	for (int j = 0; j < param.size(); j += 2) {
		double alpha = param[j];
		double theta = param[j+1];
		double x = theta / T;
		H += 3 * alpha * x / (exp(x) - 1) * T;
	}
	H *= CONST_R;
	// b) Extra terms part
	int nextraterms = extraTerms.size();
	for (int j = 0; j < nextraterms; j++) {
		H += extraTerms[j].HFuncEval(T);
	}
	return H;
}

//std::vector <double> DHFunc(std::vector <double> T1, std::vector <double> T2);
/**
 * \brief H(T2)-H(T1) function implementation (internal parameteres used)
 * \param T1 Lower temperature, K
 * \param T2 Higher temperatre, K
 */
double TDPEFunc::DHFunc(double T1, double T2)
{
	return HFunc(T2) - HFunc(T1);
}

/**
 * \brief Generalized implementation of (Cp(T), S(T)-S(0) and H(T)-H(0) functions) that use
 * internal parameters. 
 * \param dataType  Used data type
 * \param T Temperature, K
 */
std::vector <double> TDPEFunc::YFunc(TDDataType dataType, std::vector <double> T)
{
	std::vector <double> y;
	switch (dataType) {
	case DATATYPE_CP:
		y = CpFunc(T);
		break;
	case DATATYPE_S:
		y = SFunc(T);
		break;
	case DATATYPE_H:
		y = HFunc(T);
		break;
	}	
	return y;
}

// Used only by TDPEFunc::EvalDebyeTInPoint
void func_debyep(double *TD, double *res, int m, int n, void *adata)
{
	double *vals = (double *) adata;
	double T0 = vals[0], natoms = vals[1], Cp = vals[2];
	
	double Cp_D = natoms * 3.0*CONST_R*c_debye_func(TD[0] / T0);
	res[0] = Cp_D - Cp;
}

/**
 * \brief Evaluates Debye temperature for one point with T0 using Cp curve
 *
 * This function uses the next estimation of Debye temperature TD:
 * 1) Calculate Cp(T0)
 * 2) Find TD for Debye model by means of the least squares method using Debye function
 */
double TDPEFunc::EvalDebyeTInPoint(double T0, int natoms)
{
	double inpdata[] = {T0, (double) natoms, this->CpFunc(T0)};
	double TD = 100.0, TD_lb = 0.0, TD_ub  = 1e6, covar;
	double opts[LM_OPTS_SZ] = {LM_INIT_MU, 1e-15, 1e-15, 1e-20, LM_DIFF_DELTA};
	double info[LM_INFO_SZ];

	dlevmar_bc_dif(
		func_debyep, &TD, NULL, 1, 1,
		&TD_lb, &TD_ub, NULL, // Constraints (box)
		500, // Itmax
		opts, info, NULL, &covar,
		(void *) inpdata);
	// TODO: ERROR CHECK!!!
	return TD;
}




// =============== LSQContext methods ========================/

LSQContext::LSQContext()
{
	useConstraints = false;
}


/**
 * \brief Return a point with a required index
 */
Point LSQContext::GetPoint(int index)
{
	return data[index];
}

/**
 * \brief Return the number of points
 */
int LSQContext::GetNumOfPoints()
{
	return data.size();
}

/**
 * \brief Return the number of points that are included into baseline
 * (i.e. part of data that doesn't contain 2nd order phase transitions)
 */
int LSQContext::GetNumOfBaseLinePoints()
{
	return baseLineData.size();
}

/**
 * \brief Add new point to the class
 */
void LSQContext::PushPoint(Point point)
{
	data.push_back(point);
}


bool LSQContext::IsSeriesEmpty(std::string serID)
{
	for (int i = 0; i < data.size(); i++) {
		if (data[i].getSeriesID() == serID) {
			return false;
		}
	}
	return true;
}

bool LSQContext::IsSeriesEmpty(std::string serID, TDDataType ftype)
{
	for (int i = 0; i < data.size(); i++) {
		if (data[i].getSeriesID() == serID && data[i].dataType == ftype) {
			return false;
		}
	}
	return true;
}


/**
 * \brief  Returns a vector that contains all T values for 
 * all points that are kept inside the class
 */
std::vector <double> LSQContext::GetXVector(std::string serID)
{
	std::vector <double> x;
	bool allSeries = (serID == "-");
	
	for (int i = 0; i < data.size(); i++) {
		if (allSeries || data[i].getSeriesID() == serID) {
			x.push_back(data[i].x);
		}
	}
	return x;
}

/**
 * \brief  Returns a vector that contains T values for the
 * selected experimental data subset (Cp, H or S).
 *
 * \param dataType  Type of data used for selection of experimental data subset
 */
std::vector <double> LSQContext::GetXVector(TDDataType dataType, std::string serID)
{
	std::vector <double> x;
	bool allSeries = (serID == "-");

	for (int i = 0; i < data.size(); i++) {
		if (data[i].dataType == dataType && (allSeries || data[i].getSeriesID() == serID)) {
			x.push_back(data[i].x);
		}
	}
	return x;
}

/**
 * \brief  Returns a vector that contains all T1 values for 
 * all points that are kept inside the class
 */
std::vector <double> LSQContext::GetX1Vector(std::string serID)
{
	std::vector <double> x;
	bool allSeries = (serID == "-");
	
	for (int i = 0; i < data.size(); i++) {
		if (allSeries || data[i].getSeriesID() == serID) {
			x.push_back(data[i].x1);
		}
	}
	return x;
}

/**
 * \brief  Returns a vector that contains T1 values for the
 * selected experimental data subset (Cp, H or S).
 *
 * \param dataType  Type of data used for selection of experimental data subset
 */
std::vector <double> LSQContext::GetX1Vector(TDDataType dataType, std::string serID)
{
	std::vector <double> x;
	bool allSeries = (serID == "-");
	
	for (int i = 0; i < data.size(); i++) {
		if (data[i].dataType == dataType && (allSeries || data[i].getSeriesID() == serID)) {
			x.push_back(data[i].x1);
		}
	}
	return x;
}

/**
 * \brief  Returns a vector that contains all T2 values for 
 * all points that are kept inside the class
 */
std::vector <double> LSQContext::GetX2Vector(std::string serID)
{
	std::vector <double> x;
	bool allSeries = (serID == "-");
	
	for (int i = 0; i < data.size(); i++) {
		if (allSeries || data[i].getSeriesID() == serID) {
			x.push_back(data[i].x2);
		}
	}
	return x;
}

/**
 * \brief  Returns a vector that contains T2 values for the
 * selected experimental data subset (Cp, H or S).
 *
 * \param dataType  Type of data used for selection of experimental data subset
 */
std::vector <double> LSQContext::GetX2Vector(TDDataType dataType, std::string serID)
{
	std::vector <double> x;
	bool allSeries = (serID == "-");

	for (int i = 0; i < data.size(); i++) {
		if (data[i].dataType == dataType && (allSeries || data[i].getSeriesID() == serID)) {
			x.push_back(data[i].x2);
		}
	}
	return x;
}


/**
 * \brief  Returns a vector that contains all values (Cp, H, S) for 
 * all points that are kept inside the class
 */
std::vector <double> LSQContext::GetYVector(std::string serID)
{
	std::vector <double> y;
	bool allSeries = (serID == "-");

	for (int i = 0; i < data.size(); i++) {
		if (allSeries || data[i].getSeriesID() == serID) {
			y.push_back(data[i].y);
		}
	}
	return y;
}

/**
 * \brief  Returns a vector that contains values (Cp, H or S) for the
 * selected experimental data subset.
 *
 * \param dataType  Type of data used for selection of experimental data subset
 */
std::vector <double> LSQContext::GetYVector(TDDataType dataType, std::string serID)
{
	std::vector <double> y;
	bool allSeries = (serID == "-");

	for (int i = 0; i < data.size(); i++) {
		if (data[i].dataType == dataType && (allSeries || data[i].getSeriesID() == serID)) {
			y.push_back(data[i].y);
		}
	}
	return y;
}

/**
 * \brief Returns residuals vectors using experimental data inside the class
 * and user-supplied functions.
 *
 * \param tdf  User-supplied functions
 *
 * This function automatically takes into account data types (Cp, H, S) and
 * the type of error (absolute or residual).
 */
std::vector <double> LSQContext::GetResVector(TDPEFunc &tdf)
{
	std::vector <Point> &dat = baseLineData;
	std::vector <double> res(dat.size());

	for (int i = 0; i < dat.size(); i++) {
		// a) Calculated value
		double ycalc = 0.0;
		switch (dat[i].dataType) {
		case DATATYPE_CP:
			ycalc = tdf.CpFunc(dat[i].x);
			break;
		case DATATYPE_S:
			ycalc = tdf.SFunc(dat[i].x);
			break;
		case DATATYPE_H:
			ycalc = tdf.HFunc(dat[i].x2) - tdf.HFunc(dat[i].x1);
			break;
		}
		// b) weight
		double w = dat[i].w;
		if (dat[i].isRelDev) {
			w /= dat[i].y;
		}
		// c) calculate residual
		res[i] = (ycalc - dat[i].y) * w;
	}
	return res;
}

/**
 * \brief Returns Jacobian (matrix) in the form of vector
 *
 * It uses the next order of elements:
 *
 * [&part;Y<sub>0</sub>/&part;&beta;, &part;Y<sub>1</sub>/&part;&beta;, ...]
 */
std::vector <double> LSQContext::GetJVector(TDPEFunc &tdf)
{
	std::vector <Point> &dat = baseLineData;
	int npoints = dat.size();
	int nparams = tdf.GetNumOfParams();
	std::vector <double> J(npoints * nparams);
	double *Jptr = &J[0];
	double der2[nparams]; 
	double der1[nparams];

	
	for (int i = 0; i < npoints; i++) {
		// a) Calculated value
		switch (dat[i].dataType) {
		case DATATYPE_CP:
			tdf.dCpdBFunc(dat[i].x, Jptr);
			break;
		case DATATYPE_S:
			tdf.dSdBFunc(dat[i].x, Jptr);
			break;
		case DATATYPE_H:
			tdf.dHdBFunc(dat[i].x2, der2);
			tdf.dHdBFunc(dat[i].x1, der1);
			for (int j = 0; j < nparams; j++) {
				Jptr[j] = der2[j] - der1[j];
			}
			break;
		}
		// b) weight
		double w = dat[i].w;
		if (dat[i].isRelDev) {
			w /= dat[i].y;
		}
		for (int j = 0; j < nparams; j++) {
			Jptr[j] *= w;
		}
		// Go to the next line
		Jptr += nparams;
	}
	return J;
}


/**
 * \brief  Runs parameters optimization procedure using
 * the least squares method. It doesn't require an initial approximation
 */
LSQResult LSQContext::DataFit(bool optExTerms)
{
	std::vector <double> param0;
	param0.push_back(0.1);
	param0.push_back(10);
	return DataFitAddTerms(param0, optExTerms);
}

/**
 * \brief  Runs parameters optimization procedure with addition of extra
 * statistically significant terms
 */
LSQResult LSQContext::DataFitAddTerms(std::vector <double> param, bool optExTerms)
{
	LSQResult res, resPrev;
	bool isSignificant, isFirst = true;
	// Initial approximation
	res.SetParam(param);
	// Parameters search cycle
	do {
		res = DataFit(res.GetParam(), optExTerms);
		isSignificant = res.AreAllParamsSignificant();
		// If parameters are significant --- update initial approximation
		if (isSignificant) {
			resPrev = res;
			res.AddTerm(1e-3, 1);
		} else if (isFirst) {
			resPrev = res;
		}
		// All other iterations will not be first
		isFirst = false;
	} while (isSignificant);
	return resPrev;
}


/**
 * \brief  Runs parameters optimization procedure using
 * the least squares method and user-supplied initial approximation.
 * \param param Initial approximation
 * \param optExTerms Optimize/not optimize extra terms parameters
 */
LSQResult LSQContext::DataFit(std::vector <double> param, bool optExTerms, bool optBaseLine)
{
	LSQResult ans;
	int nparams = param.size();
	int npoints = this->data.size();
	ans.numOfPoints = npoints;
	std::vector <double> param_out = param;
	std::vector <double> covar(nparams * nparams);

	double opts[LM_OPTS_SZ] = {LM_INIT_MU, 1e-15, 1e-15, 1e-20, LM_DIFF_DELTA};
	double info[LM_INFO_SZ];

	InitBaseLineData();
	std::vector <Point> &data = this->baseLineData;
	
	// Function returns residuals vector to levmar library	
	auto func = [] (double *param, double *res, int m, int n, void *adata) -> void {
		LSQContext *lsq = (LSQContext *) adata;
		TDPEFunc tdf(param, m);
	
		std::vector <double> resv = lsq->GetResVector(tdf);
		std::copy(resv.begin(), resv.end(), res);
	};

	// Function returns residuals vector Jacobian to levmar library
	auto jacf = [] (double *param, double *J, int m, int n, void *adata) -> void {
		LSQContext *lsq = (LSQContext *) adata;
		TDPEFunc tdf(param, m);
	
		std::vector <double> Jv = lsq->GetJVector(tdf);
		std::copy(Jv.begin(), Jv.end(), J);	
	};
	// levmar library call
	if (optBaseLine) {
		if (useConstraints) { // Constrained optimization
			std::vector <double> param_lb(nparams), param_ub(nparams);
			for (int i = 0; i < param.size(); i += 2) {
				param_lb[i] = 0;   param_ub[i] = 5000; // alpha
				param_lb[i+1] = 0; param_ub[i+1] = 100000; // theta
			}
			dlevmar_bc_der(
				func, jacf, &param_out[0], NULL, param.size(), data.size(),
				&param_lb[0], &param_ub[0], NULL, // Constraints (box)
				500, // Itmax
				opts, info, NULL, &covar[0],
			(void *) this);
		} else { // Unconstrained optimization (can give negative values of parameters)	
			dlevmar_der(
				func, jacf, &param_out[0], NULL, param.size(), data.size(),
				500, // Itmax
				opts, info, NULL, &covar[0],
				(void *) this);
		}
		ans.SetExitCode(info[6]);
		// Save parameters and covariance matrix
		ans.SetParam(param_out);
		ans.SetCovar(covar);
	} else {
		// Use non-optimized parameters
		param_out = param;
		ans.SetParam(param_out, *this);
	}
	// Evaluate extra terms if required
	if (optExTerms) {
		// Real numeric evaluation
		std::vector <ExtraTerm> termsBuf;
		// a) Create copies and optimize parameters
		for (int i = 0; i < extraTerms.size(); i++) {
			termsBuf.push_back(extraTerms[i]);
			termsBuf[i].DataFit(this->data, ans);
		}
		// b) Write optimized parameters to the answer
		for (int i = 0; i < termsBuf.size(); i++) {
			ans.extraTerms.push_back(termsBuf[i]);
		}
	} else {
		// Just copy extra terms without optimization
		std::vector <ExtraTerm> termsBuf;
		for (int i = 0; i < extraTerms.size(); i++) {
			termsBuf.push_back(extraTerms[i]);
			// TODO: CONFIDENCE INTERVALS ESTIMATOR!!!
		}
		for (int i = 0; i < extraTerms.size(); i++) {
			ans.extraTerms.push_back(termsBuf[i]);
		}
	}
	// Estimate absolute, relative and maximal deviations for all variables
	ans.devCp = ans.GetYDeviations(*this, DATATYPE_CP);
	ans.devH  = ans.GetYDeviations(*this, DATATYPE_H);
	ans.devS  = ans.GetYDeviations(*this, DATATYPE_S);
	return ans;
}

struct seriesIDCustomLess {
	inline bool operator() (const Point &a, const Point &b) const {
		return ( (strcmp(a.seriesID, b.seriesID) < 0) ? true : false );
	}
};

struct seriesIDCustomIsEq {
	inline bool operator() (const Point &a, const Point &b) const {
		return ( (strcmp(a.seriesID, b.seriesID) == 0) ? true : false );
	}
};



std::vector <std::string> LSQContext::GetSeriesIDList(bool returnEmptySeries, TDDataType ftype)
{
	std::vector <std::string> idList;
	std::vector <Point> d = data;

	
	std::sort(d.begin(), d.end(), seriesIDCustomLess());
	d.erase(std::unique(d.begin(), d.end(), seriesIDCustomIsEq()), d.end());

	if (ftype == DATATYPE_UNKNOWN) {
		for (int i = 0; i < d.size(); i++) {
			if (returnEmptySeries || !IsSeriesEmpty(d[i].getSeriesID())) {
				idList.push_back(d[i].getSeriesID());
			}
		}
	} else {
		for (int i = 0; i < d.size(); i++) {
			if (returnEmptySeries || !IsSeriesEmpty(d[i].getSeriesID(), ftype)) {
				idList.push_back(d[i].getSeriesID());
			}
		}		
	}

	return idList;
}

void LSQContext::InitBaseLineData()
{
	int npoints = data.size();
	int nterms = extraTerms.size();
	baseLineData.clear();	
	for (int i = 0; i < npoints; i++) {
		int isBaseLine = true;
		Point &p = data[i];
		for (int j = 0; j < nterms; j++) {
			ExtraTerm &term = extraTerms[j];
			// TODO: ENTHALPY ISSUES!
			if ( (p.dataType == DATATYPE_CP && (term.Tmin <= p.x && p.x <= term.Tmax)) ||
			     (false && p.dataType == DATATYPE_H && (
				     (p.x1 <= term.Tmin && term.Tmin <= p.x2) ||
				     (p.x1 <= term.Tmax && term.Tmax <= p.x2) ))
			) {
				isBaseLine = false;
			}
		}
		if (isBaseLine) {
			baseLineData.push_back(p);
		}
	}
}



LSQResult::LSQResult()
{
	sumsqr = 0.0;
	sigma2 = 0.0;	
	lsqExitCode = 0;
}

/**
 * \brief  Loads covariance matrix and sets coefficients uncertainties
 * (standard deviations and confidence intervals)
 *
 * Requires data array, i.e experimental points, to be loaded
 * (to estimate t distribution quantile)
 */
bool LSQResult::SetCovar(const std::vector <double> &covar)
{
	// Check matrix size and copy it
	int nparams = param.size();
	if (covar.size() != nparams * nparams) {
		return false;
	}
	this->covar = covar;	
	// Estimate coefficients standard deviations
	double tinv = tdistr_inv(0.975, numOfPoints - nparams); // For 95% conf int (t-distr)
	sparam.clear();
	dparam.clear();
	for (int i = 0; i < nparams; i++) {
		int ind = i * nparams + i; // Main diagonal
		double sp = sqrt(this->covar[ind]);
		sparam.push_back(sp);
		dparam.push_back(sp * tinv);
	}
	return true;
}

/**
 * \brief  Checks if all parameters values (param vector) are positive
 */
bool LSQResult::isParamPositive()
{
	int isPositive = true;
	for (int i = 0; i < param.size(); i++) {
		if (param[i] < 0) {
			isPositive = false;
			break;
		}
	}
	return isPositive;
}


/**
 * \brief  Return parameters standard deviations
 */
std::vector <double> LSQResult::GetParamStd()
{
	return sparam;
}

/**
 * \brief  Returns parameters confidence intervals (95%, t-distr)
 */
std::vector <double> LSQResult::GetParamConfInt()
{
	return dparam;
}



/**
 * \brief  Loads model parameter and resets all deviations
 */
void LSQResult::SetParam(std::vector <double> param)
{
	int nparams = param.size();
	sumsqr = 0.0;
	sigma2 = 0.0;
	this->param = param;
	this->sparam = std::vector <double>(nparams); 
	this->dparam = std::vector <double>(nparams);
	// Fill coefficients approximations with zeros
	for (int i = 0; i < nparams; i++) {
		this->sparam[i] = 0.0;
		this->dparam[i] = 0.0;
	}
	// Fill standard deviations with zeros
	this->covar = std::vector <double>(nparams*nparams);
	for (int i = 0; i < nparams * nparams; i++) {
		this->covar[i] = 0.0;
	}
}

void LSQResult::AddTerm(double alpha, double theta)
{
	param.push_back(alpha);
	param.push_back(theta);
	SetParam(param);
}


/**
 * \brief  Loads model parameters and tries to restore confidence intervals
 *
 * If lsqCont contains some points deviations, covariance matrix and
 * confidence intervals will be calculated
 */
void LSQResult::SetParam(std::vector <double> param, LSQContext &lsqCont)
{
	int nparams = param.size();
	// Set parameter (with initialization of some structures)
	SetParam(param);
	lsqCont.InitBaseLineData();
	TDPEFunc baseLineFunc(&param[0], nparams);
	// Estimate absolute, relative and maximal deviations for all variables
	devCp = GetYDeviations(lsqCont, DATATYPE_CP);
	devH  = GetYDeviations(lsqCont, DATATYPE_H);
	devS  = GetYDeviations(lsqCont, DATATYPE_S);
	// Estimate Jacobian
	int npoints = lsqCont.GetNumOfBaseLinePoints();
	numOfPoints = npoints;
	if (npoints > 0) {
		// a) Sum of squares
		std::vector <double> res = lsqCont.GetResVector(baseLineFunc);
		for (int i = 0; i < npoints; i++) {
			sumsqr += res[i] * res[i];
		}		
		sigma2 = sumsqr / (npoints - nparams);
		// b) Jacobian and covariance matrix
		std::vector <double> Jvec = lsqCont.GetJVector(baseLineFunc);
		std::vector <double> JTJ(nparams * nparams);
		for (int i = 0; i < nparams; i++) {
			for (int j = 0; j < nparams; j++) {
				double *Jelem = &JTJ[i*nparams + j];
				*Jelem = 0;
				for (int k = 0; k < npoints; k++) {
					*Jelem += Jvec[k*nparams + i] * Jvec[k*nparams + j];
				}
			}
		}
		dlevmar_covar(&JTJ[0], &covar[0], sumsqr, nparams, npoints);
		SetCovar(covar);
	}
}


/**
 * \brief  Calculates standard deviations for chosen function type (Cp(T), S(T), H(T))
 * and for the given T value. This standard deviation can be used to calculate
 * a confidence interval of the function.
 *
 * The next formula is used to calculate standard deviation of the function:
 *
 * <i>s</i>(<i>Y</i>) = &sigma;                                                                                                                                                                                                                                                                                                                                                                                      (<i>J</i><sub>x</sub>'(<i>J</i>'<i>J</i>)<sup>-1</sup> <i>J</i><sub>x</sub> )<sup>0.5</sup>
 *                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                
 * where <i>J</i><sub>x</sub> is column vector and &sigma;(<i>J</i>'<i>J</i>)<sup>-1</sup> is covariance matrix
 */
double LSQResult::FuncConfIntStd(double T, TDDataType ftype)
{
	// Formula without HTML formatting
	// (s(Y) = sqrt(Jx' * sigma^2*inv(J'*J) * Jx)

	int m = param.size();
	double *C = &covar[0];
	std::vector <double> Jx(m);
	
	switch (ftype) {
	case DATATYPE_CP:
		dCpdBFunc(T, &Jx[0]);
		break;
	case DATATYPE_S:
		dSdBFunc(T, &Jx[0]);
		break;
	case DATATYPE_H:
		dHdBFunc(T, &Jx[0]);
		break;
	}

	// A = Jx' * sigma^2*inv(J'*J) product	
	std::vector <double> A(m);
	for (int i = 0; i < m; i++) {
		A[i] = 0.0;
		for (int j = 0; j < m; j++) {
			A[i] += Jx[j] * C[i*m + j]; // covar is symmetric
		}
	}
	// T * Jx product
	double ci_std = 0.0;
	for (int i = 0; i < m; i++) {
		ci_std += A[i] * Jx[i];
	}
	ci_std = sqrt(ci_std);


	// Part 2: extra terms conf.int.
	double ci_std_extra = 0.0;
	switch (ftype) {
	case DATATYPE_CP:
		for (int j = 0; j < extraTerms.size(); j++) {
			ExtraTerm &et = extraTerms[j];
			if (et.Tmin <= T && T <= et.Tmax) {
				double s = et.FuncConfIntStd(T, ftype);
				ci_std_extra += s*s;
			}
		}
		break;
	case DATATYPE_H:
	case DATATYPE_S:
		for (int j = 0; j < extraTerms.size(); j++) {
			ExtraTerm &et = extraTerms[j];
			if (et.Tmin <= T) {
				double s = et.FuncConfIntStd((T > et.Tmax) ? et.Tmax : T, ftype);
				ci_std_extra += s*s;
			}
		}
		break;
	}
	ci_std_extra = sqrt(ci_std_extra);

	return sqrt(ci_std*ci_std + ci_std_extra*ci_std_extra);
}

/**
 * \brief Returns precalculated deviations for a requested subset of
 * experimental data (Cp(T), S(T)-S(0), H(T)-H(0)
 * 
 * \param ftype  Type of experimental data
 */
LSQDevInfo LSQResult::GetYDeviations(TDDataType ftype)
{
	switch (ftype) {
	case DATATYPE_CP:
		return devCp;
		break;
	case DATATYPE_H:
		return devH;
		break;
	case DATATYPE_S:
		return devS;
		break;
	}
	return LSQDevInfo();
}

/**
 * \brief Calculate deviations for Y value given by user in LSQContext class example
 * described by the current parameters set (inside curent LSQResult class example).
 * For internal use only.
 */
LSQDevInfo LSQResult::GetYDeviations(LSQContext &lsqCont, TDDataType ftype)
{
	// Get all required data
	std::vector <double> T = lsqCont.GetXVector(ftype);
	std::vector <double> Y = lsqCont.GetYVector(ftype);
	std::vector <double> T1 = lsqCont.GetX1Vector(ftype);
	std::vector <double> T2 = lsqCont.GetX2Vector(ftype);
	
	if (T.size() == 0 || Y.size() == 0) {
		return LSQDevInfo();
	}
	std::vector <double> Ycalc = (ftype != DATATYPE_H) ? YFunc(ftype, T) : DHFunc(T1, T2);
	// Calculate deviations
	LSQDevInfo devs(Ycalc, Y, (ftype != DATATYPE_H) ? T : T2, 0); // We don't use parameters size (n^(-1) in std formula!)
	return devs;
}

int LSQResult::GetExitCode()
{
	return lsqExitCode;
}

bool LSQResult::SetExitCode(int exitCode)
{
	if (0 <= exitCode && exitCode <= 7) {
		lsqExitCode = exitCode;
		return true;
	} else {
		return false;
	}
}


std::string LSQResult::GetExitMessage()
{
	return std::string(lsqExitMessages[lsqExitCode]);
}

bool LSQResult::IsExitCodeGood()
{
	if (lsqExitCode == 0 || lsqExitCode == 1 ||
	    lsqExitCode == 2 || lsqExitCode == 6) {
		return true;
	} else {
		return false;
	}
}

/**
 * \brief  Check if all parameters are statistically significant
 * (t-distribution based criteria)
 */
bool LSQResult::AreAllParamsSignificant()
{
	bool isSignificant = true;
	for (int i = 0; i < param.size(); i++) {
		if (param[i] <= dparam[i]) {
			isSignificant = false;
			break;
		}
	}
	return isSignificant;	
}


const char *LSQResult::lsqExitMessages[8] = {
	"", // Code 0
	"Stopped by small gradient J^T e", // Code 1
	"Stopped by small Dp", // Code 2
	"Maximum of number of iterations exceeded", // Code 3
	"Singular matrix. Restart from current p with increased mu ", // Code 4
	"No further error reduction is possible. Restart with increased mu", // Code 5
	"Stopped by small ||e||_2", // Code 6
	"Stopped by invalid (i.e. NaN or Inf) \"func\" values. This is a user error" // Code 7
};


/**
 * \brief Initializes the class example by invalid values (to show to a user
 * that it requires filling with valid data)
 */
LSQDevInfo::LSQDevInfo()
{
	maxAbsDev = -1.0;
	maxRelDev = -1.0;
	maxAbsDevX = -1.0;
	maxRelDevX = -1.0;
	stdAbsDev = -1.0;
	stdRelDev = -1.0;
}

/**
 * \brief Initializes the class example using data series
 * \param Ycalc  Calculated values  of dependent variable
 * \param Yobs   Observed values of dependend variable
 * \param X      Value of independent variable
 * \param nparams Number of model parameters (required to calculate degrees of freedom
 * as npoints - nparam where npoints is the number of points
 */
LSQDevInfo::LSQDevInfo(std::vector <double> &Ycalc, std::vector <double> Yobs, std::vector <double> &X, int nparams)
{
	LSQDevInfo::Initialize(Ycalc, Yobs, X, nparams);
}

bool LSQDevInfo::Initialize(std::vector <double> &Ycalc, std::vector <double> Yobs, std::vector <double> &X, int nparams)
{
	// Calculate deviations
	int npoints = Ycalc.size();
	if (npoints != Yobs.size() || npoints != X.size()) {
		return false;
	}
	double sumSqrAbs = 0.0, sumSqrRel = 0.0;
	maxAbsDev = 0.0;
	maxRelDev = 0.0;
	for (int i = 0; i < npoints; i++) {
		double absDev = fabs(Ycalc[i] - Yobs[i]);
		double relDev = fabs(absDev / Yobs[i]);
		sumSqrAbs += absDev * absDev;
		sumSqrRel += relDev * relDev;
		if (absDev > maxAbsDev) {
			maxAbsDev = absDev;
			maxAbsDevX = X[i];
		}
		if (relDev > maxRelDev) {
			maxRelDev = relDev;
			maxRelDevX = X[i];
		}
	}
	stdAbsDev = sqrt(sumSqrAbs / (npoints - nparams));
	stdRelDev = sqrt(sumSqrRel / (npoints - nparams));
	return true;
}

bool LSQDevInfo::IsInitialized()
{
	if (maxAbsDev == -1.0 && stdAbsDev == -1.0) {
		return false;
	} else {
		return true;
	}
}


////////////////
ExtraTerm::ExtraTerm(ExtraTermType type)
{
	SetTermType(type);
}

double ExtraTerm::SFuncDefTerm(double T, double *p, double *c, double Tmin, double Tmax)
{
	if (T < Tmin) {
		return 0.0;		
	} else {
		double T1 = (Tmin - 10 > 1e-3) ? (Tmin - 10) : 1e-3; 
		double T2 = (T > Tmax + 10) ? (Tmax + 10) : T;
		double Ttr = c[0], A = p[0], B;
		// Left part (below Ttr)
		B = p[1]*(p[2] + 1);
		double S = A * (eminusx_ei(B*std::min(T2, Ttr)) - eminusx_ei(B*T1)*exp(B*(T1-Ttr)) );
		if (T2 > Ttr) { // Right part (above Ttr)					
			B = p[1]*(p[2] - 1);
			S += A * (eminusx_ei(B*T2)*exp(B*(T2 - Ttr)) - eminusx_ei(B*Ttr) );
		}
		return S * CONST_R;
	}	
}

double ExtraTerm::HFuncDefTerm(double T, double *p, double *c, double Tmin, double Tmax)
{
	if (T < Tmin) {
		return 0.0;
	} else {
		double T1 = (Tmin - 10 > 1e-3) ? (Tmin - 10) : 1e-3; 
		double T2 = (T > Tmax + 10) ? (Tmax + 10) : T;
		double Ttr = c[0], A = p[0], B, H = 0;
				
		if (T2 <= Ttr) {
			B = p[1]*(p[2] + 1);
			H = A * ( exp(B*(T2 - Ttr)) - exp(B*(T1 - Ttr)) ) / B;
		} else {
			// Left part (below Ttr)
			B = p[1]*(p[2] + 1);
			H = A * (1 - exp(B*(T1 - Ttr)) ) / B;
			// Right part (above Ttr)					
			B = p[1]*(p[2] - 1);
			H += A * ( exp(B*(T2 - Ttr)) - 1 ) / B;
		}
		return H * CONST_R;
	}
}


void ExtraTerm::SetTermType(ExtraTermType type)
{
	this->type = type;
	lsqparam.clear();
	fixparam.clear();
	lsqparam_names.clear();
	fixparam_names.clear();
	if (type == EXTRATERM_DEFAULT) {
		Tmin = 280; Tmax = 320;
		// Default parameters values
		lsqparam.push_back(0.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
		lsqparam.push_back(0.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
		lsqparam.push_back(0.0); lsqparam_lb.push_back(-1.0); lsqparam_ub.push_back(1.0);
		fixparam.push_back(300.0);
		// Parameters names
		lsqparam_names.push_back("b1");
		lsqparam_names.push_back("b2");
		lsqparam_names.push_back("b3");
		fixparam_names.push_back("Ttr");		
		// Some default functions	
		CpFunc = [] (double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax) -> double
		{
			double dT = T - c[0];
			return p[0] * exp(p[1]*(p[2]*dT - fabs(dT))) * CONST_R;
		};
		HFunc = [] (double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax) -> double
		{
			return ExtraTerm::HFuncDefTerm(T, &p[0], &c[0], Tmin, Tmax);
		};
		// Build SFunc using numerical integration
		SFunc = [] (double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax) -> double
		{
			return ExtraTerm::SFuncDefTerm(T, &p[0], &c[0], Tmin, Tmax);
		};		
		// Add H and S functions using numerical integration	
		//InitHSFuncs();
		// Add Cp, H and S functions derivatives using numerical differentiation
		InitDerFuncs();
	} else if (type == EXTRATERM_TWOPEAKS) {
		Tmin = 280; Tmax = 330;
		// Default parameters values
		lsqparam.push_back(0.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
		lsqparam.push_back(0.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
		lsqparam.push_back(0.0); lsqparam_lb.push_back(-1.0); lsqparam_ub.push_back(1.0);
		lsqparam.push_back(0.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
		lsqparam.push_back(0.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
		lsqparam.push_back(0.0); lsqparam_lb.push_back(-1.0); lsqparam_ub.push_back(1.0);
		fixparam.push_back(300.0);
		fixparam.push_back(310.0);
		// Parameters names
		lsqparam_names.push_back("d1(1)");
		lsqparam_names.push_back("d2(1)");
		lsqparam_names.push_back("d3(1)");
		lsqparam_names.push_back("d1(2)");
		lsqparam_names.push_back("d2(2)");
		lsqparam_names.push_back("d3(2)");
		fixparam_names.push_back("Ttr1");
		fixparam_names.push_back("Ttr2");
		// Some default functions	
		CpFunc = [] (double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax) -> double
		{
			double dT1 = T - c[0];
			double dT2 = T - c[1];
			return p[0] * exp(p[1]*(p[2]*dT1 - fabs(dT1))) * CONST_R + 
			       p[3] * exp(p[4]*(p[5]*dT2 - fabs(dT2))) * CONST_R;
		};
		HFunc = [] (double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax) -> double
		{
			return ExtraTerm::HFuncDefTerm(T, &p[0], &c[0], Tmin, Tmax) + 
				ExtraTerm::HFuncDefTerm(T, &p[3], &c[1], Tmin, Tmax);
		};
		
		SFunc = [] (double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax) -> double
		{
			return ExtraTerm::SFuncDefTerm(T, &p[0], &c[0], Tmin, Tmax) + 
				ExtraTerm::SFuncDefTerm(T, &p[3], &c[1], Tmin, Tmax);
		};
		// Add H and S functions using numerical integration
		//InitHSFuncs();
		// Add Cp, H and S functions derivatives using numerical differentiation
		InitDerFuncs();
	}
	// TODO: ADD MULTIPLE TYPE SUPPORT! (E.G. SHOTTKY)
	InitParamStd(); // st.dev and conf.int. reset
}

void ExtraTerm::InitHSFuncs()
{
	// Build HFunc using numerical integration
	HFunc = [CpFunc] (double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax) -> double
	{
		auto ff = [CpFunc, &p, &c, &Tmin, &Tmax] (double Ti) -> double { return CpFunc(Ti, p, c, Tmin, Tmax);};
		if (T < Tmin) {
			return 0.0;		
		} else {
			double T1 = (Tmin - 50 > 1e-3) ? (Tmin - 50) : 1e-3; 
			double T2 = (T > Tmax + 50) ? (Tmax + 50) : T;
			return int_adaptive_simpson(ff, T1, T2);
		}
	};
	// Build SFunc using numerical integration
	SFunc = [CpFunc] (double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax) -> double
	{
		auto ff = [CpFunc, &p, &c, &Tmin, &Tmax] (double Ti) -> double { return CpFunc(Ti, p, c, Tmin, Tmax) / Ti;};
		if (T < Tmin) {
			return 0.0;
		} else {
			double T1 = (Tmin - 50 > 1e-3) ? (Tmin - 50) : 1e-3; 
			double T2 = (T > Tmax + 50) ? (Tmax + 50) : T;
			return int_adaptive_simpson(ff, T1, T2);
		}
	};
}

void ExtraTerm::InitDerFuncs()
{
	// Build functions derivatives
	dCpdBFunc = [CpFunc] (double T, double *J, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax) -> double
	{
		for (int i = 0; i < p.size(); i++) {
			double dB = 1e-3;
			std::vector <double> pB = p; pB[i] = pB[i] + dB;
			std::vector <double> pA = p; pA[i] = pA[i] - dB;
			J[i] = (CpFunc(T, pB, c, Tmin, Tmax) - CpFunc(T, pA, c, Tmin, Tmax)) / (2*dB);
		}
		return 0.0;
	};

	dSdBFunc = [SFunc] (double T, double *J, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax) -> double
	{
		for (int i = 0; i < p.size(); i++) {
			double dB = 1e-3;
			std::vector <double> pB = p; pB[i] = pB[i] + dB;
			std::vector <double> pA = p; pA[i] = pA[i] - dB;
			J[i] = (SFunc(T, pB, c, Tmin, Tmax) - SFunc(T, pA, c, Tmin, Tmax)) / (2*dB);
		}
		return 0.0;
	};
	
	dHdBFunc = [HFunc] (double T, double *J, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax) -> double
	{
		for (int i = 0; i < p.size(); i++) {
			double dB = 1e-3;
			std::vector <double> pB = p; pB[i] = pB[i] + dB;
			std::vector <double> pA = p; pA[i] = pA[i] - dB;
			J[i] = (HFunc(T, pB, c, Tmin, Tmax) - HFunc(T, pA, c, Tmin, Tmax)) / (2*dB);
		}
		return 0.0;
	};
}

void ExtraTerm::InitParamStd()
{
	slsqparam.clear();
	dlsqparam.clear();
	for (int i = 0; i < lsqparam.size(); i++) {
		slsqparam.push_back(0.0);
		dlsqparam.push_back(0.0);
	}
}


bool ExtraTerm::SetLSQParam(std::vector <double> &p)
{
	if (p.size() == lsqparam.size()) {
		lsqparam = p;
		return true;
	} else {
		return false;
	}
}

std::vector <double> ExtraTerm::GetLSQParam()
{
	return lsqparam;
}

std::vector <std::string> ExtraTerm::GetLSQParamNames()
{
	return lsqparam_names;
}

std::vector <double> ExtraTerm::GetFixParam()
{
	return fixparam;
}

std::vector <std::string> ExtraTerm::GetFixParamNames()
{
	return fixparam_names;
}

std::vector <double> ExtraTerm::GetLSQParamStd()
{
	return slsqparam;
}

std::vector <double> ExtraTerm::GetLSQParamCi()
{
	return dlsqparam;
}

std::vector <double> ExtraTerm::GetLSQParamLB()
{
	return lsqparam_lb;
}

std::vector <double> ExtraTerm::GetLSQParamUB()
{
	return lsqparam_ub;
}


bool ExtraTerm::SetFixParam(std::vector <double> &c)
{
	if (c.size() == fixparam.size()) {
		fixparam = c;
		return true;
	} else {
		return false;
	}
}

bool ExtraTerm::DataFit(std::vector <Point> &data, TDPEFunc &baseLine)
{
	// Preselect experimental data (with baseline subtraction)
	for (int i = 0; i < data.size(); i++) {
		Point p = data[i];
		if (p.dataType == DATATYPE_CP && Tmin <= p.x && p.x <= Tmax) {
			p.y = p.y - baseLine.CpFunc(p.x);
			etdata.push_back(p);
		}
	}
	// Check data size
	if (etdata.size() <= lsqparam.size()) {
		return false; // Not enough experimental data (or too many model parameters)
	}
	// lambda for optimization procedure
	auto minfunc = [] (double *param, double *res, int m, int n, void *adata) -> void
	{
		ExtraTerm *term = (ExtraTerm *) adata;
		// Create lsqparam
		std::vector <double> lsqparam(m);
		for (int i = 0; i < m; i++) {
			lsqparam[i] = param[i];
		}
		// Calculate residuals
		for (int i = 0; i < n; i++) {
			// a) calculated value (for Cp only!)
			double T = term->etdata[i].x;
			double ycalc = term->CpFunc(T, lsqparam, term->fixparam, term->Tmin, term->Tmax);
			double yexp = term->etdata[i].y;
			// b) weight
			double w = term->etdata[i].w;
			if (term->etdata[i].isRelDev) {
				w /= yexp;
			}
			// c) calculate residual
			res[i] = (ycalc - yexp) * w;			
		}
	};
	// Optimization procedure
	int nparams = lsqparam.size();
	int npoints = etdata.size();
	std::vector <double> lsqparam_out = lsqparam;
	covar = std::vector <double> (nparams * nparams);

	double opts[LM_OPTS_SZ] = {LM_INIT_MU, 1e-15, 1e-15, 1e-20, LM_DIFF_DELTA};
	double info[LM_INFO_SZ];

	dlevmar_bc_dif(
		minfunc, &lsqparam_out[0], NULL, nparams, npoints,
		&lsqparam_lb[0], &lsqparam_ub[0], NULL, // Constraints (box)
		500, // Itmax
		opts, info, NULL, &covar[0],
		(void *) this);
	lsqparam = lsqparam_out;
	// Calculate residuals
	residuals = std::vector<double>(npoints);
	minfunc(&lsqparam[0], &residuals[0], nparams, npoints, (void *) this);
	// Calculate coefficients confidence intervals
	slsqparam.clear();
	dlsqparam.clear();
	double tinv = tdistr_inv(0.975, npoints - nparams); // For 95% conf int (t-distr)
	for (int i = 0; i < nparams; i++) {
		int ind = i * nparams + i; // Main diagonal
		double sp = sqrt(covar[ind]);
		slsqparam.push_back(sp);
		dlsqparam.push_back(sp * tinv);
	}
	// TODO: ERROR CHECK!
	return true;
}

ExtraTermType ExtraTerm::GetType()
{
	return type;
}

std::string ExtraTerm::GetTypeAsString()
{
	std::string typeStr;
	switch (type) {
	case EXTRATERM_DEFAULT:
		typeStr = "DEFAULT"; break;
	case EXTRATERM_TWOPEAKS:
		typeStr = "TWOPEAKS"; break;
	}
	return typeStr;
}


int ExtraTerm::GetNumOfPoints()
{
	return etdata.size();
}

std::vector <double> ExtraTerm::GetResVector()
{
	return residuals;
}

std::vector <double> ExtraTerm::GetTVector()
{
	std::vector <double> Tvec(etdata.size());
	for (int i = 0; i < etdata.size(); i++) {
		Tvec[i] = etdata[i].x;
	}
	return Tvec;
}

std::vector <double> ExtraTerm::GetYVector()
{
	std::vector <double> Yvec(etdata.size());
	for (int i = 0; i < etdata.size(); i++) {
		Yvec[i] = etdata[i].y;
	}
	return Yvec;
}

std::vector <bool> ExtraTerm::GetIsRelDevVector()
{
	std::vector <bool> isRelDev(etdata.size());
	for (int i = 0; i < etdata.size(); i++) {
		isRelDev[i] = etdata[i].isRelDev;
	}
	return isRelDev;
}


double ExtraTerm::CpFuncEval(double T)
{
	return CpFunc(T, lsqparam, fixparam, Tmin, Tmax);
}

double ExtraTerm::HFuncEval(double T)
{
	return HFunc(T, lsqparam, fixparam, Tmin, Tmax);
}

double ExtraTerm::SFuncEval(double T)
{
	return SFunc(T, lsqparam, fixparam, Tmin, Tmax);
}

void ExtraTerm::dCpdBFuncEval(double T, double *J)
{
	dCpdBFunc(T, J, lsqparam, fixparam, Tmin, Tmax);
}

void ExtraTerm::dHdBFuncEval(double T, double *J)
{
	dHdBFunc(T, J, lsqparam, fixparam, Tmin, Tmax);
}

void ExtraTerm::dSdBFuncEval(double T, double *J)
{
	dSdBFunc(T, J, lsqparam, fixparam, Tmin, Tmax);
}


double ExtraTerm::FuncConfIntStd(double T, TDDataType ftype)
{

	// Formula without HTML formatting
	// (s(Y) = sqrt(Jx' * sigma^2*inv(J'*J) * Jx)
	
	int m = lsqparam.size();
	double *C = &covar[0];
	std::vector <double> Jx(m);
	
	switch (ftype) {
	case DATATYPE_CP:
		dCpdBFuncEval(T, &Jx[0]);
		break;
	case DATATYPE_S:
		dSdBFuncEval(T, &Jx[0]);
		break;
	case DATATYPE_H:
		dHdBFuncEval(T, &Jx[0]);
		break;
	}

	// A = Jx' * sigma^2*inv(J'*J) product	
	std::vector <double> A(m);
	for (int i = 0; i < m; i++) {
		A[i] = 0.0;
		for (int j = 0; j < m; j++) {
			A[i] += Jx[j] * C[i*m + j]; // covar is symmetric
		}
	}
	// T * Jx product
	double ci_std = 0.0;
	for (int i = 0; i < m; i++) {
		ci_std += A[i] * Jx[i];
	}
	ci_std = sqrt(ci_std);
	return ci_std;
}

