/**
 * \file 
 * \brief Contains classes that implement non-linear regression
 *
 * This file contains wxWidgets-independent implementation of non-linear
 * regression for Cp(T), S(T), H(T)-H(0) experimental data description.
 * It is based on levmar library.
 *
 * \author Alexey L. Voskov (alvoskov@gmail.com)
 * \copyright (C) 2016-2017 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#ifndef __DATAFIT_H
#define __DATAFIT_H

/** CODATA 2014 molar gas constant */
#define CONST_R 8.3144598

/** Thermochemical calorie (in J) */
#define CONST_CAL 4.184

/** Mathematical constants */
#define EULER_GAMMA 0.577215664901532860606512

/** Some useful statistical functions **/
double tdistr_cdf(double x, int f);
double tdistr_inv(double p, double f);
double eminusx_ei(double x);

/** Thermodynamic data type: Cp, S-S(0) or H-H(0) */
enum TDDataType {DATATYPE_CP, DATATYPE_S, DATATYPE_H, DATATYPE_UNKNOWN};

/**
 * \brief Data point of Cp(T), S(T) or H(T)-H(0) experimental data with
 * data type, statistical weights and type of deviation
 *
 */
class Point
{
public:
	Point();
	double x; /*!< T, K */
	double x1; /*!< T1, K for H(T2)-H(T1) heat content */
	double x2; /*!< T2, K for H(T2)-H(T1) heat content */
	double y; /*!< Cp, S-S(T) or H-H(0) value (J,mol,K-based units) */
	bool isRelDev; /*!< true/false --- use relative/absolute deviations */
	double w; /*!< Statistical weight */
	TDDataType dataType; /*!< Data type of y value */
	char seriesID[16]; /*!< Data series ID (name) */

	void setSeriesID(const char *id);
	std::string getSeriesID();
};


/** Extra term types */
enum ExtraTermType {EXTRATERM_DEFAULT, EXTRATERM_TWOPEAKS};

class TDPEFunc;
/**
 * \brief  Contains extra Cp term (e.g. for lambda-transitions, Schottky anomaly etc.)
 */
class ExtraTerm {
public:
	ExtraTerm(ExtraTermType = EXTRATERM_DEFAULT);
	void SetTermType(ExtraTermType type);

	double Tmin; /*!< Minimal temperature for extra term */
	double Tmax; /*!< Maximal temperature for extra term */
	std::vector <Point> etdata; /*!< Data for extra term parameters approximation (with subtracted baseline) */
	
	bool SetLSQParam(std::vector <double> &p);
	bool SetFixParam(std::vector <double> &c);
	std::vector <double> GetLSQParam();
	std::vector <std::string> GetLSQParamNames();
	std::vector <double> GetFixParam();
	std::vector <std::string> GetFixParamNames();
	std::vector <double> GetLSQParamStd();
	std::vector <double> GetLSQParamCi();
	std::vector <double> GetLSQParamLB();
	std::vector <double> GetLSQParamUB();
	int GetNumOfPoints();
	std::vector <double> GetResVector();
	std::vector <double> GetTVector();
	std::vector <double> GetYVector();
	std::vector <bool> GetIsRelDevVector();

	bool DataFit(std::vector <Point> &data, TDPEFunc &baseLine);

	double FuncConfIntStd(double T, TDDataType ftype);
		
	double CpFuncEval(double T);
	double HFuncEval(double T);
	double SFuncEval(double T);

	void dCpdBFuncEval(double T, double *J);
	void dHdBFuncEval(double T, double *J);
	void dSdBFuncEval(double T, double *J);

	ExtraTermType GetType();
	std::string GetTypeAsString();
private:
	ExtraTermType type;
	std::vector <double> lsqparam; /*!< Model parameters for extra term (to optimize by LSQ) */
	std::vector <double> slsqparam; /*!< lsqparam standard deviations */
	std::vector <double> dlsqparam; /*!< lsqparam confidence intervals */
	std::vector <double> lsqparam_lb; /*!< Model parameters for LSQ extra term (lower bounds) */
	std::vector <double> lsqparam_ub; /*!< Model parameters for LSQ extra term (upper bounds) */
	std::vector <double> fixparam; /*!< Model parameters for extra terms (fixed) */
	std::vector <double> residuals; /*!< Residuals vector (result of DataFit) */
	std::vector <double> covar; /*!< Covariance matrix from levmar library */

	std::vector <std::string> fixparam_names; /*!< Names (string) of fixed parameters */
	std::vector <std::string> lsqparam_names; /*!< Names (string) of LSQ parameters */
	
	std::function <double (double, std::vector <double> &, std::vector <double> &, double Tmin, double Tmax) -> double > CpFunc;
	std::function <double (double, std::vector <double> &, std::vector <double> &, double Tmin, double Tmax) -> double > HFunc;
	std::function <double (double, std::vector <double> &, std::vector <double> &, double Tmin, double Tmax) -> double > SFunc;
	
	std::function <double (double, double *J, std::vector <double> &, std::vector <double> &, double Tmin, double Tmax) -> double > dCpdBFunc;
	std::function <double (double, double *J, std::vector <double> &, std::vector <double> &, double Tmin, double Tmax) -> double > dHdBFunc;
	std::function <double (double, double *J, std::vector <double> &, std::vector <double> &, double Tmin, double Tmax) -> double > dSdBFunc;


	void InitHSFuncs();
	void InitDerFuncs();
	void InitParamStd();

	static double SFuncDefTerm(double T, double *p, double *c, double Tmin, double Tmax);
	static double HFuncDefTerm(double T, double *p, double *c, double Tmin, double Tmax);
	
};


/**
 * \brief Contains parameters set and functions for calculation of
 * Cp(T), S(T)-S(0) and H(T)-H(0) values for user-defined temperatures.
 *
 * This class is an implementation of Einstein-Plank functions suggested
 * by Voronin et al. for description of thermodynamic functions. CODATA 2014
 * universal gas constant value (R = 8.3144598) is used.
 *
 * Voronin G.F., Kutsenok I.B. Universal Method for Approximating the Standard
 * Thermodynamic Functions of Solids // J. Chem. Eng. Data. 2013. 58(7). 2083--2094.
 * http://dx.doi.org/10.1021/je400316m 
 */
class TDPEFunc {
public:
	// Functions
	std::vector <double> CpFunc(std::vector <double> T);
	double CpFunc(double T);
	void dCpdBFunc(double T, double *J);
	
	std::vector <double> SFunc(std::vector <double> T);
	double SFunc(double T);
	void dSdBFunc(double T, double *J);
	
	std::vector <double> HFunc(std::vector <double> T);
	double HFunc(double T);
	void dHdBFunc(double T, double *J);

	std::vector <double> DHFunc(std::vector <double> T1, std::vector <double> T2);
	double DHFunc(double T1, double T2);
	
	std::vector <double> YFunc(TDDataType dataType, std::vector <double> T);
	// Functions for working with parameters	
	std::vector <double> GetParam();
	std::pair < std::vector <double>, std::vector<double> > GetAlphaTheta();
	int GetNumOfParams();
	virtual void AddTerm(double alpha, double theta);
	// Constructors
	TDPEFunc();
	TDPEFunc(double *param, int nparams);
	// Some extra functions
	double EvalDebyeTInPoint(double T0, int natoms);

	// Extra terms with optimized parameters values
	std::vector <ExtraTerm> extraTerms; /*!< Extra terms with optimized parameters values */
protected:
	/**
         * \brief Model parameres
         *
         * This vector keeps all model parameters (&alpha;<sub>i</sub> and &theta;<sub>i</sub>)
         * using the next order in memory (&beta; is the notation for param vector elements):
         *
         * &beta;[0] = &alpha;[0], &beta;[1] = &theta;[0], &beta;[2] = &alpha;[1], &beta;[3] = &theta;[1] etc.
         */
	std::vector <double> param;
};


/**
 * \brief  Contains deviations calculated for a set of parameters obtained
 * by the least squares method.
 */
class LSQDevInfo {
public:
	double maxAbsDev; /*!< Maximal absolute deviation */
	double maxRelDev; /*!< Maximal relative deviation */
	double maxAbsDevX; /*!< X for maximal absolute deviation */
	double maxRelDevX; /*!< X for maximal relative deviation */
	double stdAbsDev; /*!< Standard absolute deviation */
	double stdRelDev; /*!< Standard relative deviation */
	LSQDevInfo();
	LSQDevInfo(std::vector <double> &Ycalc, std::vector <double> Yobs, std::vector <double> &X, int nparams);
	bool Initialize(std::vector <double> &Ycalc, std::vector <double> Yobs, std::vector <double> &X, int nparams);
	bool IsInitialized();
};

class LSQContext;

/**
 * \brief  Contains results of the least squares method run.
 * Optimized parameters, their standard deviations, different kinds of deviations
 * for experimental data can be obtained by the user. This class also calculates
 * confidence intervals for functions.
 */
class LSQResult : public TDPEFunc {
public:
	// Variables
	LSQDevInfo devCp; /*!< Deviations for Cp(T) data */
	LSQDevInfo devH; /*!< Deviations for H(T)-H(0) data */
	LSQDevInfo devS; /*!< Deviations for S(T)-S(0) data */
	int numOfPoints; /*!< Number of points used in the optimization */
	// Raw LSQ output	
	double sumsqr; /*!< Sum of squares (raw LSQ output) */
	double sigma2; /*!< Variance of predicted variable (Cp) (raw LSQ output) */
	// Functions
	LSQResult();
	void SetParam(std::vector <double> param);
	void SetParam(std::vector <double> param, LSQContext &lsqCont);	
	void AddTerm(double alpha, double theta);
	std::vector <double> GetParamStd();
	std::vector <double> GetParamConfInt();
	bool SetCovar(const std::vector <double> &covar);
	bool isParamPositive();
	double FuncConfIntStd(double T, TDDataType ftype);
	LSQDevInfo GetYDeviations(TDDataType ftype);
	LSQDevInfo GetYDeviations(LSQContext &lsqCont, TDDataType ftype);
	// Functions for returning optimization result
	bool AreAllParamsSignificant();
	// Functions for working with LSQ result status
	int GetExitCode();
	bool SetExitCode(int exitCode);
	std::string GetExitMessage();
	bool IsExitCodeGood();
private:
	std::vector <double> sparam; /*!< Std.dev of parameters */
	std::vector <double> dparam; /*!< 95% confidence intervals (t-distr) of parameters */
	std::vector <double> covar; /*!< Covariance matrix */

	int lsqExitCode;
	static const char *lsqExitMessages[8];
};


/**
 * \brief Contains an input set of experimental data and subroutions
 * to optimize model parameters. This class actively uses levmar library
 * for its work.
 */
class LSQContext {
public:
	std::vector <Point> data; /*!< Experimental data set */
	std::vector <Point> baseLineData; /*!< Experimental data set with excluded anomalies */
	bool useConstraints; /*!< Use parameters constraints */

	std::vector <ExtraTerm> extraTerms;
	
	LSQContext();
	Point GetPoint(int index);
	int GetNumOfPoints();
	int GetNumOfBaseLinePoints();
	void PushPoint(Point point);
	bool IsSeriesEmpty(std::string serID);
	bool IsSeriesEmpty(std::string serID, TDDataType ftype);
	std::vector <double> GetXVector(std::string serID = std::string("-"));
	std::vector <double> GetXVector(TDDataType dataType, std::string serID = std::string("-"));
	std::vector <double> GetX1Vector(std::string serID = std::string("-"));
	std::vector <double> GetX1Vector(TDDataType dataType, std::string serID = std::string("-"));
	std::vector <double> GetX2Vector(std::string serID = std::string("-"));
	std::vector <double> GetX2Vector(TDDataType dataType, std::string serID = std::string("-"));
	std::vector <double> GetYVector(std::string serID = std::string("-"));
	std::vector <double> GetYVector(TDDataType dataType, std::string serID = std::string("-"));
	std::vector <double> GetResVector(TDPEFunc &tdf);
	std::vector <double> GetJVector(TDPEFunc &tdf);
	std::vector <std::string> GetSeriesIDList(bool returnEmptySeries = true, TDDataType ftype = DATATYPE_UNKNOWN);

	LSQResult DataFit(bool optExTerms = true);
	LSQResult DataFit(std::vector <double> param, bool optExTerms = true, bool optBaseLine = true);
	LSQResult DataFitAddTerms(std::vector <double> param, bool optExTerms = true);

	void InitBaseLineData();
};


#endif
