/**
 * \file
 * \copyright (C) 2017 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */


/*
 * Debye function implementation
 */
#include <functional>
#include <math.h>
#include <float.h>
#include <stdio.h>

static double c[] = {
       -0.0005952380952380953, // c*y^4, |d|=0.0482
        1.102292768959436e-05, // c*y^6, |d|=0.00804
       -2.254689754689755e-07, // c*y^8, |d|=0.00148
        4.817713151046485e-09, // c*y^10, |d|=0.000284
       -1.056838027737499e-10, // c*y^12, |d|=5.62e-05
        2.361624093650237e-12, // c*y^14, |d|=1.13e-05
       -5.352126783667236e-14, // c*y^16, |d|=2.3e-06
        1.226580293753978e-15, // c*y^18, |d|=4.75e-07
       -2.836785258988776e-17, // c*y^20, |d|=9.89e-08
        6.610803394032275e-19, // c*y^22, |d|=2.07e-08
       -1.550496076201391e-20, // c*y^24, |d|=4.38e-09
        3.656593489271863e-22, // c*y^26, |d|=9.29e-10
       -8.664694284229884e-24, // c*y^28, |d|=1.98e-10
        2.061774956670621e-25, // c*y^30, |d|=4.25e-11
       -4.924106287604745e-27, // c*y^32, |d|=9.12e-12
        1.179869574822864e-28, // c*y^34, |d|=1.97e-12
         -2.8353807235887e-30, // c*y^36, |d|=4.26e-13
        6.831756773484179e-32, // c*y^38, |d|=9.23e-14
       -1.650015638860905e-33, // c*y^40, |d|=2.01e-14
         3.99378117498809e-35, // c*y^42, |d|=4.37e-15
        -9.68588224128696e-37, // c*y^44, |d|=9.54e-16
        2.353321301298482e-38, // c*y^46, |d|=2.09e-16
       -5.727266513321583e-40, // c*y^48, |d|=4.57e-17
        1.395988892943782e-41, // c*y^50, |d|=1e-17
       -3.407496404629496e-43, // c*y^52, |d|=2.2e-18
        8.328436970865744e-45, // c*y^54, |d|=4.84e-19
       -2.038105264041832e-46, // c*y^56, |d|=1.07e-19
        4.993315994542405e-48, // c*y^58, |d|=2.35e-20
       -1.224668646781844e-49, // c*y^60, |d|=5.19e-21
        3.006672016261761e-51, // c*y^62, |d|=1.15e-21
};

static int c_num = sizeof(c) / sizeof(double);

double dfunc3_smallx(double x)
{
	double f = 1 - 0.375*x + 0.05*x*x;
	for (int k = c_num - 1; k >= 0; k--) {
		f += c[k] * pow(x, 4 + 2*k);
	}
	return f;
}

double dfunc3_largex(double x)
{
	double f = 0;
	for (int k = 20; k >= 1; k--) {
		double xk = x*k;
		f += 3*exp(-xk)/k*(1 + 3/xk*(1 + 2/xk*(1 + 1/xk)));
	}
	return pow(M_PI, 4)/15 * 3*pow(x, -3) - f;
}


double debye_func_3(double x)
{
	return (x > 3) ? dfunc3_largex(x) : dfunc3_smallx(x);
}


/*
 * C/(N*3*R) function
 */
double c_debye_func(double x)
{
	if (x >= DBL_EPSILON) {
		return 4*debye_func_3(x) - 3*x/(exp(x) - 1);
	} else {
		return 4*debye_func_3(x) - 3/(1 + x/2);
	}
}



double simpson(std::function <double(double)> f, double a, double b)
{
	double c = a + 0.5*(b - a);
	return (b - a) / 6 * (f(a) + 4*f(c) + f(b));
}


double simpson_recursive(std::function <double(double)> f, double a, double b, double Sab, double eps, int iter)
{
	double c = a + 0.5*(b - a);
	double Sac = simpson(f, a, c);
	double Scb = simpson(f, c, b);
	double Sab_new = Sac + Scb;
	if (iter > 10 || Sab == INFINITY || Sab == -INFINITY || Sab != Sab) {
		return Sab;
	}
	if (fabs(Sab_new - Sab) / 15 < eps || eps < DBL_MIN) {
		return Sab + (Sab_new - Sab);
	} else {
		return simpson_recursive(f, a, c, Sac, eps / 2, iter + 1) + 
		       simpson_recursive(f, c, b, Scb, eps / 2, iter + 1);
	}
}

/*
 * Adaptive Simpson
 */
double int_adaptive_simpson(std::function <double(double)> f, double a, double b)
{
	double Sab0 = simpson(f, a, b);
	return simpson_recursive(f, a, b, Sab0, DBL_EPSILON, 0);
}


double CpTnorm_func(double Tnorm)
{
	if (Tnorm < DBL_EPSILON) {
		return 0;
	} else {
		return c_debye_func(1 / Tnorm) / Tnorm;
	}
}

/*
 * S / (N*3*R) function
 * Tnorm = T / TD
 */
double S_debye_func(double Tnorm)
{
	return int_adaptive_simpson(CpTnorm_func, 0, Tnorm);
}
