/**
 * \file
 * \copyright (C) 2016-2017 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#include <vector>
#include <functional>
#include "wx/wx.h"
#include "wx/grid.h"
#include "wx/bookctrl.h"
#include "wx/textfile.h"
#include "datafit.h"
#include "plotcanvas.h"
#include "cp.h"
#include "graph.h"
#include "tdtable.h"
#include "xyplots.h"
#include "misc.h"
#include "dfunc3.h"


/**
 * \brief Window with low-temperature Debye-Huckel correction term
 */
class DebyeCorrResultWindow : public wxFrame
{
public:
	DebyeCorrResultWindow(wxFrame *parent, TDPEFunc *tdpefunc, double natoms, double T0);
	void OnSize(wxSizeEvent &event);
	void OnClose(wxCloseEvent &event);
private:
	double natoms;
	double T0;
	double TD;
	TDPEFunc tdpefunc;
	
	wxTextCtrl *txtNAtoms;
	wxTextCtrl *txtT0;
	wxTextCtrl *txtTD;
	wxTextCtrl *txtSE;
	wxTextCtrl *txtSD;
	wxTextCtrl *txtDS;

	PlotCanvas *pl;
	
	
	wxDECLARE_EVENT_TABLE();
};

wxBEGIN_EVENT_TABLE(DebyeCorrResultWindow, wxFrame)
	EVT_SIZE(DebyeCorrResultWindow::OnSize)
	EVT_CLOSE(DebyeCorrResultWindow::OnClose)
wxEND_EVENT_TABLE()


DebyeCorrResultWindow::DebyeCorrResultWindow(wxFrame *parent, TDPEFunc *tdpefunc, double natoms, double T0) :
	wxFrame(parent, wxID_ANY, "Low-T entropy correction result", wxPoint(50, 50), wxSize(700, 500))
{
	this->tdpefunc = *tdpefunc;
	this->natoms = natoms;
	this->T0 = T0;
	this->TD = TD;
	
	wxPanel *panel = new wxPanel(this, -1);
	wxPoint ctlPos(10, 25);
	// Labels
	// a) For textboxes
	wxStaticText *lblNAtoms = new wxStaticText(panel, wxID_ANY, "Number of atoms", wxPoint(10, 25), wxSize(150, 25));
	wxStaticText *lblT0 = new wxStaticText(panel, wxID_ANY, "T0 (lowest T), K", wxPoint(10, 60), wxSize(150, 25));
	wxStaticText *lblTD = new wxStaticText(panel, wxID_ANY, "TD (debye T), K", wxPoint(10, 95), wxSize(150, 25));
	wxStaticText *lblSE = new wxStaticText(panel, wxID_ANY, "S (E.-P.), J/(mol*K)", wxPoint(10, 130), wxSize(150, 25));
	wxStaticText *lblSD = new wxStaticText(panel, wxID_ANY, "S (Debye), J/(mol*K)", wxPoint(10, 165), wxSize(150, 25));
	wxStaticText *lblDS = new wxStaticText(panel, wxID_ANY, "SD - SE, J/(mol*K)", wxPoint(10, 200), wxSize(150, 25));
	// b) Graph legend
	wxStaticText *lblInfo1 = new wxStaticText(panel, wxID_ANY, "Black line - E.-P. functions sum", wxPoint(10, 255), wxSize(300, 25));
	wxStaticText *lblInfo2 = new wxStaticText(panel, wxID_ANY, "Blue line - Debye function", wxPoint(10, 290), wxSize(300, 25));
	// TextBoxes for values
	txtNAtoms = new wxTextCtrl(panel, wxID_ANY, "", wxPoint(170, 25), wxSize(100, 25), wxTE_READONLY);
	txtT0 = new wxTextCtrl(panel, wxID_ANY, "", wxPoint(170, 60), wxSize(100, 25), wxTE_READONLY);
	txtTD = new wxTextCtrl(panel, wxID_ANY, "", wxPoint(170, 95), wxSize(100, 25), wxTE_READONLY);
	txtSE = new wxTextCtrl(panel, wxID_ANY, "", wxPoint(170, 130), wxSize(100, 25), wxTE_READONLY);
	txtSD = new wxTextCtrl(panel, wxID_ANY, "", wxPoint(170, 165), wxSize(100, 25), wxTE_READONLY);
	txtDS = new wxTextCtrl(panel, wxID_ANY, "", wxPoint(170, 200), wxSize(100, 25), wxTE_READONLY);
	// Evaluate corrections
	TD = this->tdpefunc.EvalDebyeTInPoint(T0, natoms);
	double SD = natoms * 3*CONST_R*S_debye_func(T0 / TD);
	double SE = this->tdpefunc.SFunc(T0);
	// Load values
	txtNAtoms->ChangeValue(wxString::Format("%.4g", natoms));
	txtT0->ChangeValue(wxString::Format("%.4g", T0));
	txtTD->ChangeValue(wxString::Format("%.4g", TD));
	txtSE->ChangeValue(wxString::Format("%.4g", SE));
	txtSD->ChangeValue(wxString::Format("%.4g", SD));
	txtDS->ChangeValue(wxString::Format("%.4g", SD - SE));
	// Plot canvas with graphs
	pl = new PlotCanvas(panel, -1, wxPoint(300, 25), wxSize(300, 300));
	pl->SetXLabel(wxT("T / K"));
	pl->SetYLabel(wxT("Cp / J/(mol\u00B7K)")); // \cdot
	
	
	std::vector <double> T, CpDv, CpEv;
	for (double Ti = 1e-5; Ti <= T0 + 1e-10; Ti += T0 * 0.001) {
		double CpDi = natoms * 3*CONST_R*c_debye_func(TD / Ti);
		T.push_back(Ti);
		CpDv.push_back(CpDi);
		
	}
	CpEv = this->tdpefunc.CpFunc(T);
	pl->Plot(T, CpEv, PlotMarker(wxColor(0,0,0), wxColor(0,0,0), LINE_SOLID, MARKER_NONE));
	pl->Plot(T, CpDv, PlotMarker(wxColor(0,0,255), wxColor(0,0,0), LINE_SOLID, MARKER_NONE));
	pl->FitAxes();
	// Set window icon
	SetIcon(wxICON(sample));
	// Set window coordinates and minimal size
	wxPoint wndPos = parent->GetPosition();
	wndPos.x += 20;
	wndPos.y += 20;
	SetPosition(wndPos);
	SetMinSize(wxSize(700, 500));
}


void DebyeCorrResultWindow::OnSize(wxSizeEvent &event)
{
        wxSize s = this->GetClientSize();
	s.SetWidth(s.GetWidth() - 320);
	s.SetHeight(s.GetHeight() - 35);
	pl->SetSize(s);
	pl->Move(300, 25);	
	event.Skip();
}


void DebyeCorrResultWindow::OnClose(wxCloseEvent &event)
{
	event.Skip();
}




//////////////////////////////////////////////////////////
enum {
	ID_BTN_EVALDEBYECORR = 1000,
};


class DebyeCorrWindow : public wxFrame
{
public:
	DebyeCorrWindow(wxFrame *parent, TDPEFunc *tdpefunc);
	void OnSize(wxSizeEvent &event);
	void OnClose(wxCloseEvent &event);
	void OnEvalDebyeCorr(wxCommandEvent &event);
private:
	wxTextCtrl *txtNAtoms;
	wxTextCtrl *txtT0;
	TDPEFunc tdpefunc;

	wxDECLARE_EVENT_TABLE();
};

wxBEGIN_EVENT_TABLE(DebyeCorrWindow, wxFrame)
	EVT_SIZE(DebyeCorrWindow::OnSize)
	EVT_CLOSE(DebyeCorrWindow::OnClose)
	EVT_BUTTON(ID_BTN_EVALDEBYECORR, DebyeCorrWindow::OnEvalDebyeCorr)
wxEND_EVENT_TABLE()


/**
 * @brief  Opens window to evaluate influence of differences between Einstein-Plank and Debye functions
 * on entropy values at low (T < 30 K) temperatures
 */
DebyeCorrWindow::DebyeCorrWindow(wxFrame *parent, TDPEFunc *tdpefunc)
	: wxFrame(parent, wxID_ANY, "Low-T entropy correction", wxPoint(50, 50), wxSize(300, 180))
{
	this->tdpefunc = *tdpefunc;
	wxPanel *panel = new wxPanel(this, -1);
	wxPoint ctlPos(10, 25);
	wxStaticText *lblNAtoms = new wxStaticText(panel, wxID_ANY, "Number of atoms", wxPoint(10, 25), wxSize(150, 25));
	wxStaticText *lblT0 = new wxStaticText(panel, wxID_ANY, "T0 (lowest T), K", wxPoint(10, 60), wxSize(150, 25));
	
	txtNAtoms = new wxTextCtrl(panel, wxID_ANY, "", wxPoint(170, 25), wxSize(100, 25));
	txtT0 = new wxTextCtrl(panel, wxID_ANY, "", wxPoint(170, 60), wxSize(100, 25));
	
	wxButton *btnEvalDebyeCorr = new wxButton(panel, ID_BTN_EVALDEBYECORR, "Evaluate", wxPoint(170,100), wxSize(100,30));
	
	// Set window icon
	SetIcon(wxICON(sample));
}


void DebyeCorrWindow::OnSize(wxSizeEvent &event)
{	
	event.Skip();
}


void DebyeCorrWindow::OnClose(wxCloseEvent &event)
{
	event.Skip();
}

///////////////////////////

enum {
	ID_BTN_BUILDDHPLOT = 1000,
	ID_TXT_DHPLOT_T0,
	ID_TXT_DHPLOT_DT0
};

/**
 * \brief  Dialog for H(T)-H(T0) and (H(T)-H(T0))/(T - T0) plots
 * parameters customization
 */
class DHPlotDialog : public wxFrame
{
public:
	DHPlotDialog(wxFrame *parent, LSQContext &lsqCont, LSQResult &lsqRes);
	void OnSize(wxSizeEvent &event);
	void OnClose(wxCloseEvent &event);
	void OnBuildDHPlot(wxCommandEvent &event);
	void OnTxtT0(wxCommandEvent &event);
	void OnTxtDT0(wxCommandEvent &event);
private:
	wxTextCtrl *txtT0;
	wxTextCtrl *txtDT0;
	wxCheckBox *chkNormed;
	LSQContext lsqCont;
	LSQResult lsqRes;
	double T0, DT0;
	
	wxDECLARE_EVENT_TABLE();
};

wxBEGIN_EVENT_TABLE(DHPlotDialog, wxFrame)
	EVT_SIZE(DHPlotDialog::OnSize)
	EVT_CLOSE(DHPlotDialog::OnClose)
	EVT_BUTTON(ID_BTN_BUILDDHPLOT, DHPlotDialog::OnBuildDHPlot)
	EVT_TEXT(ID_TXT_DHPLOT_T0, DHPlotDialog::OnTxtT0)
	EVT_TEXT(ID_TXT_DHPLOT_DT0, DHPlotDialog::OnTxtDT0)
wxEND_EVENT_TABLE()


DHPlotDialog::DHPlotDialog(wxFrame *parent, LSQContext &lsqCont, LSQResult &lsqRes)
	: wxFrame(parent, wxID_ANY, "H(T)-H(T0) plot construction", wxPoint(50, 50), wxSize(300, 220))
{
	this->lsqCont = lsqCont;
	this->lsqRes = lsqRes;
	T0 = 298.15; DT0 = 0.5;

	wxPanel *panel = new wxPanel(this, -1);
	wxStaticText *lblT0 = new wxStaticText(panel, wxID_ANY, "T0, K", wxPoint(10, 25), wxSize(150, 25));
	wxStaticText *lblDT0 = new wxStaticText(panel, wxID_ANY, wxT("\u0394T0, K"), wxPoint(10, 60), wxSize(150, 25));
	
	txtT0 = new wxTextCtrl(panel, ID_TXT_DHPLOT_T0, wxString::Format("%g", T0), wxPoint(170, 25), wxSize(100, 25));
	txtDT0 = new wxTextCtrl(panel, ID_TXT_DHPLOT_DT0, wxString::Format("%g", DT0), wxPoint(170, 60), wxSize(100, 25));
	chkNormed = new wxCheckBox(panel, wxID_ANY, "Divide by (T - T0)", wxPoint(10, 100), wxSize(200, 25));
	chkNormed->SetValue(true);
	
	wxButton *btnBuildPlot = new wxButton(panel, ID_BTN_BUILDDHPLOT, "Build plot", wxPoint(170,135), wxSize(100,30));

	// Set window icon
	SetIcon(wxICON(sample));
}

void DHPlotDialog::OnSize(wxSizeEvent &event)
{	
	event.Skip();
}


void DHPlotDialog::OnClose(wxCloseEvent &event)
{
	event.Skip();
}

void DHPlotDialog::OnBuildDHPlot(wxCommandEvent &event)
{
	wxString strTitle = wxString::Format("CpFit H(T) - H(%g) data", T0);
	DHDataWindow *dHwnd = new DHDataWindow(this, strTitle, lsqCont, lsqRes, T0, DT0, chkNormed->GetValue());
	dHwnd->Show();
}

/**
 * \brief Controls T0 input and saves successful attempts inside the object (as double value)
 */
void DHPlotDialog::OnTxtT0(wxCommandEvent &event)
{
	double T;
	if (!event.GetString().ToDouble(&T) || T < 0.0 || T > 20000.0) {
		txtT0->ChangeValue(wxString::Format("%g", T0)); // Bad input: restore old value
	} else {
		T0 = T; // Good input: remember new value
	}
}

/**
 * \brief Controls DT0 input and saves successful attempts inside the object (as double value)
 */
void DHPlotDialog::OnTxtDT0(wxCommandEvent &event)
{
	double T;
	if (!event.GetString().ToDouble(&T) || T < 0 || T > 100) {
		txtDT0->ChangeValue(wxString::Format("%g", DT0)); // Bad input: restore old value
	} else {
		DT0 = T; // Good input: remember new value
	}
}


/**
 * @brief Runs the evaluation process
 */
void DebyeCorrWindow::OnEvalDebyeCorr(wxCommandEvent &event)
{
	double natoms, T0;
	// Get values	
	wxString strNAtoms = txtNAtoms->GetValue();
	wxString strT0 = txtT0->GetValue();
	if (!strNAtoms.ToDouble(&natoms) || natoms < 1.0) {
		wxMessageBox("Invalid natoms value", "CpFit", wxICON_ERROR);
		return;
	}
	if (!strT0.ToDouble(&T0) || T0 <= 0.0 || T0 > 500.0) {
		wxMessageBox("Invalid T0 value", "CpFit", wxICON_ERROR);
		return;
	}
	// Find Debye temperature and entropy values
	DebyeCorrResultWindow *wnd = new DebyeCorrResultWindow(this, &tdpefunc, natoms, T0);
	wnd->Show(true);	
	event.Skip();
}


////////////////////////////////////////////////////////////
DevInfoPanel::DevInfoPanel(wxWindow *parent, wxPoint pos, wxSize size, wxString unit)
	: wxPanel(parent, wxID_ANY, pos, size)
{
	// Create all objects for the Cp model errors static box
	int xLblNames = 10, wLblNames = 170, wTxtDevs = 110, wLblAt = 25, wTxtK = 80;
	wxSize sizeLblNames(wLblNames, 25);
	wxSize sizeTxtDevs(wTxtDevs, 25);
	wxSize sizeLblAt(wLblAt, 25);
	wxSize sizeTxtK(wTxtK, 25);
	wxSize sizeLblK(25, 25);
	// a) Absolute deviations
	wxPoint ctlPos(xLblNames, 10);
	wxStaticText *lblStdAbsDev = new wxStaticText(this, wxID_ANY, "Std.absolute dev.:", ctlPos, sizeLblNames); // Names
	ctlPos.x += wLblNames;
	txtStdAbsDev = new wxTextCtrl(this, wxID_ANY, "", ctlPos, sizeTxtDevs); // Devs
	txtStdAbsDev->SetEditable(false);
	// b) Relative deviations
	ctlPos.x = xLblNames; ctlPos.y += 35;
	wxStaticText *lblStdRelDev = new wxStaticText(this, wxID_ANY, "Std.relative dev.(%):", ctlPos, sizeLblNames); // At
	ctlPos.x += wLblNames;
	txtStdRelDev = new wxTextCtrl(this, wxID_ANY, "", ctlPos, sizeTxtDevs); // K text
	txtStdRelDev->SetEditable(false);
	// c) Max. Abs. Deviations (with T)
	ctlPos.x = xLblNames; ctlPos.y += 35;
	wxStaticText *lblMaxAbsDev = new wxStaticText(this, wxID_ANY, "Max.absolute dev.:", ctlPos, sizeLblNames); // Names
	ctlPos.x += wLblNames;
	txtMaxAbsDev = new wxTextCtrl(this, wxID_ANY, "", ctlPos, sizeTxtDevs); // Devs
	txtMaxAbsDev->SetEditable(false);
	ctlPos.x += wTxtDevs + 5;
	wxStaticText *lblMaxAbsDevAt = new wxStaticText(this, wxID_ANY, "at", ctlPos, sizeLblAt); // At
	ctlPos.x += wLblAt;
	txtMaxAbsDevX = new wxTextCtrl(this, wxID_ANY, "", ctlPos, sizeTxtK); // K text
	txtMaxAbsDevX->SetEditable(false);
	ctlPos.x += wTxtK + 5;
	wxStaticText *lblMaxAbsDevK  = new wxStaticText(this, wxID_ANY, "K", ctlPos, sizeLblK); // K label
	// d) Max Rel. Dev (with T)
	ctlPos.x = xLblNames; ctlPos.y += 35;
	wxStaticText *lblMaxRelDev = new wxStaticText(this, wxID_ANY, "Max.relative dev.(%):", ctlPos, sizeLblNames); // Names
	ctlPos.x += wLblNames;
	txtMaxRelDev = new wxTextCtrl(this, wxID_ANY, "", ctlPos, sizeTxtDevs); // Devs
	txtMaxRelDev->SetEditable(false);
	ctlPos.x += wTxtDevs + 5;
	wxStaticText *lblMaxRelDevAt = new wxStaticText(this, wxID_ANY, "at", ctlPos, sizeLblAt); // At
	ctlPos.x += wLblAt;
	txtMaxRelDevX = new wxTextCtrl(this, wxID_ANY, "", ctlPos, sizeTxtK); // K text
	txtMaxRelDevX->SetEditable(false);
	ctlPos.x += wTxtK + 5;
	wxStaticText *lblMaxRelDevK  = new wxStaticText(this, wxID_ANY, "K", ctlPos, sizeLblK); // K label
	// e) Info about dimensions
	ctlPos.x = xLblNames; ctlPos.y += 35;
	wxStaticText *lblDimInfo = new wxStaticText(this, wxID_ANY,
		wxString::Format(L"All absolute errors are in %s", unit), ctlPos, wxSize(300, 30));
}

void DevInfoPanel::LoadLSQDevInfo(LSQDevInfo &dev)
{
	if (dev.IsInitialized()) {
		txtMaxAbsDev->ChangeValue(wxString::Format("%.4g", dev.maxAbsDev));
		txtMaxAbsDevX->ChangeValue(wxString::Format("%.4g", dev.maxAbsDevX));
		txtMaxRelDev->ChangeValue(wxString::Format("%.4g", dev.maxRelDev * 100));
		txtMaxRelDevX->ChangeValue(wxString::Format("%.4g", dev.maxRelDevX));
		txtStdAbsDev->ChangeValue(wxString::Format("%.4g", dev.stdAbsDev));
		txtStdRelDev->ChangeValue(wxString::Format("%.4g", dev.stdRelDev * 100));
	} else {
		txtMaxAbsDev->ChangeValue("");
		txtMaxAbsDevX->ChangeValue("");
		txtMaxRelDev->ChangeValue("");
		txtMaxRelDevX->ChangeValue("");
		txtStdAbsDev->ChangeValue("");
		txtStdRelDev->ChangeValue("");
	}
}


////////////////////////////////////////////////////////////
enum {
	ID_GR_PARAM = 100,
	ID_BTN_COPYPARAMS,
	
	ID_MNU_CPPLOT,
	ID_MNU_CPTPLOT,
	ID_MNU_HPLOT,
	ID_MNU_SPLOT,
	ID_MNU_DHPLOT,

	ID_MNU_CPABSRESPLOT,
	ID_MNU_CPRELRESPLOT,

	ID_MNU_HABSRESPLOT,
	ID_MNU_HRELRESPLOT,

	ID_MNU_TDTABLE,
	ID_MNU_CPRESTABLE,
	ID_MNU_HRESTABLE,

	ID_MNU_SAVEASCSV,
	ID_MNU_SAVEASHTML,

	ID_MNU_DEBYECORR
};

wxBEGIN_EVENT_TABLE(GraphWindow, wxFrame)
	EVT_SIZE(GraphWindow::OnSize)
	EVT_CLOSE(GraphWindow::OnClose)
	EVT_BUTTON(ID_BTN_COPYPARAMS, GraphWindow::OnCopyParams)

	EVT_MENU(ID_MNU_SAVEASCSV, GraphWindow::OnSaveAsCSV)
	EVT_MENU(ID_MNU_SAVEASHTML, GraphWindow::OnSaveAsHTML)
	
	EVT_MENU(ID_MNU_CPPLOT, GraphWindow::OnCpPlot)
	EVT_MENU(ID_MNU_CPTPLOT, GraphWindow::OnCpTPlot)
	EVT_MENU(ID_MNU_HPLOT, GraphWindow::OnHPlot)
	EVT_MENU(ID_MNU_SPLOT, GraphWindow::OnSPlot)
	EVT_MENU(ID_MNU_DHPLOT, GraphWindow::OnDHPlot)

	EVT_MENU(ID_MNU_TDTABLE, GraphWindow::OnShowTdTable)
	EVT_MENU(ID_MNU_CPRESTABLE, GraphWindow::OnShowCpResTable)
	EVT_MENU(ID_MNU_HRESTABLE, GraphWindow::OnShowHResTable)
	
	EVT_MENU(ID_MNU_CPABSRESPLOT, GraphWindow::OnCpAbsResPlot)
	EVT_MENU(ID_MNU_CPRELRESPLOT, GraphWindow::OnCpRelResPlot)
	
	EVT_MENU(ID_MNU_HABSRESPLOT, GraphWindow::OnHAbsResPlot)
	EVT_MENU(ID_MNU_HRELRESPLOT, GraphWindow::OnHRelResPlot)
	
	EVT_MENU(ID_MNU_DEBYECORR, GraphWindow::OnDebyeCorrection)
wxEND_EVENT_TABLE()


/**
 * \brief Initializes all controls and menus
 *
 * Note: complex interaction with GraphWindow::OnSize function by means of some sizes
 * initialized in this function.
 */
GraphWindow::GraphWindow(MainWindow *parent, const wxString &title)
	: wxFrame(parent, wxID_ANY, title, wxPoint(50, 50), wxSize(600, 650))
{
	boxParams = NULL; boxLsq = NULL;
	wxPanel *panel = new wxPanel(this, -1);
	// Create all objects in the parameters static box
	boxParams = new wxStaticBox(panel, wxID_ANY, "Model parameters", wxPoint(10, 10), wxSize(420, 230));
	nbkParams = new wxNotebook(boxParams, wxID_ANY, wxPoint(10, 10), wxSize(420, 230));
	ctlBlParams = new wxPanel(nbkParams, wxID_ANY, wxPoint(0, 0), wxSize(420,230));
	ctlExParams = new wxPanel(nbkParams, wxID_ANY, wxPoint(0, 0), wxSize(420,230));

	wxFont *fntCourier = new wxFont(10, wxFONTFAMILY_TELETYPE, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL);	
	txtExParamsReport = new wxTextCtrl(ctlExParams, wxID_ANY, "Report\nReport\n",
		wxPoint(20, 30), wxSize(10, 10), wxTE_READONLY | wxTE_MULTILINE);
	txtExParamsReport->SetFont(*fntCourier);
		
	nbkParams->AddPage(ctlBlParams, "Base line");
	nbkParams->AddPage(ctlExParams, "Extra terms");


	grPar = new wxGrid(ctlBlParams, ID_GR_PARAM, wxPoint(20, 20), wxSize(380,160));
	grPar->CreateGrid(3, 6);
	grPar->SetColLabelValue(0, wxT("\u03B1")); // alpha
	grPar->SetColLabelValue(1, wxT("\u0394(\u03B1)")); // DELTA(alpha)
	grPar->SetColLabelValue(2, wxT("s(\u03B1)")); // s(alpha)
	grPar->SetColLabelValue(3, wxT("\u03B8 / K")); // theta
	grPar->SetColLabelValue(4, wxT("\u0394(\u03B8) / K")); // DELTA(theta)
	grPar->SetColLabelValue(5, wxT("s(\u03B8) / K")); // s(theta)
	grPar->SetRowLabelSize(40);
	btnCopyParams = new wxButton(boxParams, ID_BTN_COPYPARAMS, "Use as initial approximation", wxPoint(20,190), wxSize(380,30));
	// Create all objects for the Cp model errors static box
	boxLsq = new wxStaticBox(panel, wxID_ANY, "Model errors", wxPoint(10, 250), wxSize(420, 260));
	nbkTabsDev = new wxNotebook(boxLsq, wxID_ANY, wxPoint(10, 25), wxSize(400, 210));
	CpDevPanel = new DevInfoPanel(nbkTabsDev, wxPoint(0, 0), wxSize(400, 180));
	HDevPanel = new DevInfoPanel(nbkTabsDev, wxPoint(0, 0), wxSize(400, 180), "J / mol");
	nbkTabsDev->AddPage(CpDevPanel, "Cp (heat capacity)");
	nbkTabsDev->AddPage(HDevPanel, "H(T2)-H(T1) (enthalpy)");
	// Set window icon
	SetIcon(wxICON(sample));	
	// Create user menu
	wxMenuBar *menuBar = new wxMenuBar();	
	// Create file menu
	wxMenu *menuFile = new wxMenu();
	menuFile->Append(ID_MNU_SAVEASCSV, "Save as CSV", "Save as CSV");
	menuFile->Append(ID_MNU_SAVEASHTML, "Save as HTML", "Save as HTML");
	menuBar->Append(menuFile, "File");	
	// Create plots menu
	wxMenu *menuPlots = new wxMenu();
	menuBar->Append(menuPlots, "Plots");
	// Plots menu: Cp plots submenu
	wxMenu *menuCpPlots = new wxMenu();
	menuCpPlots->Append(ID_MNU_CPPLOT, "Function", "Function");
	menuCpPlots->Append(ID_MNU_CPABSRESPLOT, "Absolute residuals", "Absolute residuals");
	menuCpPlots->Append(ID_MNU_CPRELRESPLOT, "Relative residuals", "Relative residuals");
	menuPlots->AppendSubMenu(menuCpPlots, "Cp function");
	// Plots menu: Cp/T plots submenu
	wxMenu *menuCpTPlots = new wxMenu();
	menuCpTPlots->Append(ID_MNU_CPTPLOT, "Cp/T plot", "Cp/T function");
	menuPlots->AppendSubMenu(menuCpTPlots, "Cp/T function");
	// Plots menu: H plots submenu
	wxMenu *menuHPlots = new wxMenu();
	menuHPlots->Append(ID_MNU_HPLOT, "H plot", "H plot");
	menuHPlots->Append(ID_MNU_HABSRESPLOT, "Absolute residuals", "Absolute residuals");
	menuHPlots->Append(ID_MNU_HRELRESPLOT, "Relative residuals", "Relative residuals");
	menuHPlots->Append(ID_MNU_DHPLOT, "H(T)-H(T0) plot", "H(T)-H(T0) plot");
	menuPlots->AppendSubMenu(menuHPlots, "H function");
	// Plots menu: S plots submenu
	wxMenu *menuSPlots = new wxMenu();
	menuSPlots->Append(ID_MNU_SPLOT, "S plot", "S plot");
	menuPlots->AppendSubMenu(menuSPlots, "S function");
	// Plots menu: Tables menu	
	wxMenu *menuTables = new wxMenu();
	menuTables->Append(ID_MNU_TDTABLE, "Thermodynamic functions", "Thermodynamic functions");
	wxMenu *menuResiduals = new wxMenu();
	menuResiduals->Append(ID_MNU_CPRESTABLE, "Cp residuals", "Cp residuals");
	menuResiduals->Append(ID_MNU_HRESTABLE, "H(T2)-H(T1) residuals", "H(T2)-H(T1) residuals");
	menuTables->AppendSubMenu(menuResiduals, "Residuals");
	menuBar->Append(menuTables, "Tables");
	// Evaluate menu: different numeric estimations
	wxMenu *menuEvals = new wxMenu();
	menuEvals->Append(ID_MNU_DEBYECORR, "Low-T S correction", "Low-T S correction");
	menuBar->Append(menuEvals, "Evaluate");
	// Append the whole menu to the form
	SetMenuBar(menuBar);
	// Set window coordinates and sizes
	wxPoint wndPos = parent->GetPosition();
	wndPos.x += 20;
	wndPos.y += 20;
	SetPosition(wndPos);
	SetMinSize(wxSize(600, 700));
}

/**
 * \brief Set experimental data
 */
void GraphWindow::AddData(LSQContext lsq)
{
	this->lsq = lsq;
}

/**
 * \brief Set initial approximation
 */
void GraphWindow::AddParam0(std::vector <double> alpha, std::vector <double> theta)
{
	alpha0 = alpha;
	theta0 = theta;
}

/**
 * \brief Run the least squares method
 *
 * Starts the least squares method (LSQContext::dataFit) using initial approximation
 * given by a user and loads all results to the controls of the class. The main results
 * are coefficients, their standard deviations, and absolute and relative deviations
 * of Cp data description.
 */
void GraphWindow::CurveFit(GraphWindowFitMode fitMode)
{
	std::vector <double> param;
	int nterms;

	// Run data fitting	
	// a) initialize initial approximation (if required)
	if (fitMode != FIT_VALNTERMS_AUTO) {
		nterms = alpha0.size();
		for (int i = 0; i < nterms; i++) {
			param.push_back(alpha0[i]);
			param.push_back(theta0[i]);
		}
	}
	// b) select an appropriate method
	lsq.useConstraints = false;
	switch (fitMode) {
	case FIT_VALNTERMS_AUTO:
		lsqRes = lsq.DataFit();
		if (!lsqRes.isParamPositive()) {
			lsq.useConstraints = true;
			lsqRes = lsq.DataFit();
		}
		break;
	case FIT_VALS:
		lsqRes = lsq.DataFit(param);
		if (!lsqRes.isParamPositive()) {
			lsq.useConstraints = true;
			lsqRes = lsq.DataFit(param);
		}
		break;
	case FIT_VALNTERMS_INITAPPROX:
		lsqRes = lsq.DataFitAddTerms(param);
		if (!lsqRes.isParamPositive()) {
			lsq.useConstraints = true;
			lsqRes = lsq.DataFitAddTerms(param);
		}
		break;
	case FIT_EXTERMS_ONLY:
		lsqRes = lsq.DataFit(param, true, false); // Optimize extra terms only
		break;
	case FIT_NOOPT:
		lsqRes.SetParam(param, lsq);
		lsqRes.extraTerms = lsq.extraTerms; // TODO: MORE CHECKS!!! NEW METHODS!!!
		break;
	}
	// Get parameters
	param = lsqRes.GetParam();
	std::vector <double> sparam = lsqRes.GetParamStd();
	std::vector <double> dparam = lsqRes.GetParamConfInt();
	// Warning in the case of bad optimization results
	if (!lsqRes.IsExitCodeGood()) {
		wxMessageBox(lsqRes.GetExitMessage().c_str(), "CpFit", wxICON_WARNING);
	}
	// Visualize parameters
	nterms = lsqRes.GetNumOfParams() / 2;
	grPar->DeleteRows(0, grPar->GetNumberRows());
	grPar->InsertRows(0, nterms);
	for (int i = 0; i < nterms; i++) {
		// alpha parameters
		grPar->SetCellValue(i, 0, wxString::Format("%g", param[2*i]));
		grPar->SetCellValue(i, 1, wxString::Format("%.4g", dparam[2*i]));
		grPar->SetCellValue(i, 2, wxString::Format("%.4g", sparam[2*i]));
		if (fabs(param[2*i] <= dparam[2*i])) {
			for (int j = 0; j < 3; j++) {
				grPar->SetCellTextColour(i, j, wxColour(255, 0, 0));
			}
		} else {
			for (int j = 0; j < 3; j++) {
				grPar->SetCellTextColour(i, j, wxColour(0, 0, 0));
			}
		}
		// theta parameters
		grPar->SetCellValue(i, 3, wxString::Format("%g", param[2*i + 1]));
		grPar->SetCellValue(i, 4, wxString::Format("%.4g", dparam[2*i + 1]));
		grPar->SetCellValue(i, 5, wxString::Format("%.4g", sparam[2*i + 1]));
		if (fabs(param[2*i + 1] <= dparam[2*i + 1])) {
			for (int j = 3; j < 6; j++) {
				grPar->SetCellTextColour(i, j, wxColour(255, 0, 0));
			}
		} else {
			for (int j = 3; j < 6; j++) {
				grPar->SetCellTextColour(i, j, wxColour(0, 0, 0));
			}
		}
		// Set read only cells for alpha and theta parameters
		for (int j = 0; j < 6; j++) {
			grPar->SetReadOnly(i, j);
		}
	}
	// Visualize extra terms parameters (in the form of text report)
	wxString strExParams;
	int nexterms = lsqRes.extraTerms.size();
	if (nexterms == 0) {
		strExParams = "No extra terms are present";
	} else {
		strExParams = wxString::Format("%d extra terms are present\n", nexterms);
		for (int i = 0; i < nexterms; i++) {
			ExtraTerm &term = lsqRes.extraTerms[i];
			std::vector <double> fixparam = term.GetFixParam();
			std::vector <double> lsqparam = term.GetLSQParam();
			std::vector <double> slsqparam = term.GetLSQParamStd();
			std::vector <double> dlsqparam = term.GetLSQParamCi();
			std::vector <double> Tv = term.GetTVector();
			std::vector <double> Yv = term.GetYVector();
			std::vector <double> resv = term.GetResVector();
			std::vector <bool> isrelv = term.GetIsRelDevVector();
			// General information
			strExParams += wxString::Format("TERM %d\n", i + 1);
			strExParams += "Type: " + term.GetTypeAsString() + "\n";
			strExParams += wxString::Format("Tmin = %g K, Tmax = %g K\n", term.Tmin, term.Tmax);
			// Fixed parameters info
			strExParams += wxString::Format("Fixed parameters:\n");
			strExParams += wxString::Format("  %3s %12s\n", "#", "A");
			for (int j = 0; j < fixparam.size(); j++) {
				strExParams += wxString::Format("  %3d %12g\n", j + 1, fixparam[j]);
			}
			// LSQ parameters info
			strExParams += wxString::Format("LSQ parameters:\n");
			strExParams += wxString::Format("  %3s %12s %12s %12s\n", "#", "A", "S(A)", "D(A)");
			for (int j = 0; j < lsqparam.size(); j++) {
				strExParams += wxString::Format("  %3d %12g %12g %12g\n",
					j + 1, lsqparam[j], slsqparam[j], dlsqparam[j]);
			}
			// Data series information
			strExParams += wxString::Format("Number of points: %d\n", Tv.size());
			strExParams += wxString::Format("  %3s %12s %5s %12s %12s %12s\n",
				"#", "T", "DEV", "Cp(e)-Cp(bl)", "Cp(c)-Cp(bl)", "D(Cp)");
			for (int j = 0; j < Tv.size(); j++) {
				strExParams += wxString::Format("  %3d %12g %5s %12g %12g %12g\n",
					j + 1, Tv[j], isrelv[j] ? "REL" : "ABS", Yv[j], term.CpFuncEval(Tv[j]), resv[j]);
			}
			strExParams += "\n";
		}
	}
	txtExParamsReport->SetValue(strExParams);
	// Get and visualize Cp and H deviations
	LSQDevInfo devCp = lsqRes.GetYDeviations(DATATYPE_CP);
	CpDevPanel->LoadLSQDevInfo(devCp);
	LSQDevInfo devH = lsqRes.GetYDeviations(DATATYPE_H);
	HDevPanel->LoadLSQDevInfo(devH);
}

/**
 * \brief Resizing some controls
 */
void GraphWindow::OnSize(wxSizeEvent &event)
{
        wxSize s = event.GetSize();
	if (boxParams != NULL) {
		// Resize frames (staticboxes)
		boxLsq->SetSize(s.x - 30, 250);
		boxLsq->Move(10, s.y - 320);
		boxParams->SetSize(s.x - 30, boxLsq->GetPosition().y - 20);
		wxSize nbkSize = nbkTabsDev->GetSize();
		nbkSize.x = s.x - 60;
		nbkTabsDev->SetSize(nbkSize);
		// Resize table and buttons inside parmeters staticbox
		wxSize bs = boxParams->GetSize();
		// a) button
		btnCopyParams->SetSize(bs.x - 40, 30);
		btnCopyParams->Move(20, bs.y - btnCopyParams->GetSize().y - 10);
		// b) table
		nbkParams->Move(20, 30);
		nbkParams->SetSize(bs.x - 40, btnCopyParams->GetPosition().y - 2*20);
//		grPar->Move(20, 30);
		bs = ctlBlParams->GetSize();
		grPar->SetSize(bs.x - 40, bs.y - 40);
		bs = ctlExParams->GetSize();
		txtExParamsReport->SetSize(bs.x - 40, bs.y - 40);
		ResizeWxGridCols(grPar);
	}
	
	event.Skip();
}

void GraphWindow::OnClose(wxCloseEvent &event)
{
	event.Skip();
}

/**
 * \brief Sends optimized parameters to the parent window
 *
 * Sends optimized parameters to the parent window. They will be used
 * as an initial approximation and will appear in the corresponding table
 * of the CpFit main window.
 */
void GraphWindow::OnCopyParams(wxCommandEvent &event)
{
	std::pair <std::vector <double>, std::vector <double> > pairAlphaTheta;
	pairAlphaTheta = lsqRes.GetAlphaTheta();
	MainWindow *parent = (MainWindow *) this->GetParent();
	parent->WriteAlpha(pairAlphaTheta.first);
	parent->WriteTheta(pairAlphaTheta.second);
	if (!parent->WriteExtraTermsParams(lsqRes.extraTerms)) {
		wxMessageBox("Cannot copy extra terms: sizes or types are not matching");
	}
}

/**
 * \brief Shows a window with a table of tabulated thermodynamic functions
 */
void GraphWindow::OnShowTdTable(wxCommandEvent &event)
{
	TdTableWindow *tdtab = new TdTableWindow(this, &lsqRes, "CpFit thermodynamic functions table");
	tdtab->Show();
}

/**
 * \brief Shows a window with residuals (experimental Cp vs calculated Cp)
 */
void GraphWindow::OnShowCpResTable(wxCommandEvent &event)
{
	ResTableWindow *restab = new ResTableWindow(this, DATATYPE_CP, lsq, lsqRes, wxString("CpFit residuals table"));
	restab->Show();
}

/**
 * \brief Shows a window with residuals (experimental H(T2)-H(T1) vs calculated H(T2)-H(T1))
 */
void GraphWindow::OnShowHResTable(wxCommandEvent &event)
{
	ResTableWindow *restab = new ResTableWindow(this, DATATYPE_H, lsq, lsqRes, wxString("CpFit residuals table"));
	restab->Show();
}

/**
 * \brief Shows Cp(T) plot with experimental points and calculated curve
 */
void GraphWindow::OnCpPlot(wxCommandEvent &event)
{
	YDataWindow *cpwnd = new YDataWindow(this, "CpFit Cp Data", lsq, lsqRes, DATATYPE_CP);
	cpwnd->Show();
}

/**
 * \brief Shows Cp(T)/T plot with experimental points and calculated curve
 *
 * Cp(T)/T terms graphs are also present (colored dashed lines)
 */
void GraphWindow::OnCpTPlot(wxCommandEvent &event)
{
	CpTDataWindow *cpwnd = new CpTDataWindow(this, "CpFit Cp/T Data", lsq, lsqRes);
	cpwnd->Show();
}

/**
 * \brief Shows H(T)-H(0) plot with calculated curve
 */
void GraphWindow::OnHPlot(wxCommandEvent &event)
{
	YDataWindow *cpwnd = new YDataWindow(this, "CpFit H Data", lsq, lsqRes, DATATYPE_H);
	cpwnd->Show();
}

/**
 * \brief Shows H(T)-H(T0) plot with calculated curve and experimental points
 */
void GraphWindow::OnDHPlot(wxCommandEvent &event)
{
	DHPlotDialog *dlg = new DHPlotDialog(this, lsq, lsqRes);
	dlg->Show(true);
}

/**
 * \brief Shows S(T)-S(0) plot with calculated curve
 */
void GraphWindow::OnSPlot(wxCommandEvent &event)
{
	YDataWindow *cpwnd = new YDataWindow(this, "CpFit S Data", lsq, lsqRes, DATATYPE_S);
	cpwnd->Show();
}


/**
 * \brief Shows Cp absolute residuals
 */
void GraphWindow::OnCpAbsResPlot(wxCommandEvent &event)
{
	ResDataWindow *reswnd = new ResDataWindow(this, "CpFit Cp absolute residuals", lsq, lsqRes, DATATYPE_CP, false);
	reswnd->Show();
}

/**
 * \brief Shows Cp relative residuals
 */
void GraphWindow::OnCpRelResPlot(wxCommandEvent &event)
{
	ResDataWindow *reswnd = new ResDataWindow(this, "CpFit Cp relative residuals", lsq, lsqRes, DATATYPE_CP, true);
	reswnd->Show();
}

/**
 * \brief Shows H absolute residuals
 */
void GraphWindow::OnHAbsResPlot(wxCommandEvent &event)
{
	ResDataWindow *reswnd = new ResDataWindow(this, "CpFit H absolute residuals", lsq, lsqRes, DATATYPE_H, false);
	reswnd->Show();
}

/**
 * \brief Shows H relative residuals
 */
void GraphWindow::OnHRelResPlot(wxCommandEvent &event)
{
	ResDataWindow *reswnd = new ResDataWindow(this, "CpFit H relative residuals", lsq, lsqRes, DATATYPE_H, true);
	reswnd->Show();
}


/**
 * \brief Exports parameters and their confidence intervals into CSV file
 */
void GraphWindow::OnSaveAsCSV(wxCommandEvent &event)
{
	wxFileDialog openFileDialog(this, wxT("Save image"), "", "",
                       "csv files (*.csv)|*.csv", wxFD_SAVE);
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		wxString fileName = openFileDialog.GetPath();
		wxTextFile csvFile;
		if (csvFile.Create(fileName)) {
			std::vector <double> param = lsqRes.GetParam();
			std::vector <double> sparam = lsqRes.GetParamStd();
			std::vector <double> dparam = lsqRes.GetParamConfInt();
			csvFile.AddLine("alpha;dalpha;salpha;theta;dtheta;stheta");
			for (int i = 0; i < lsqRes.GetNumOfParams(); i+=2) {
				csvFile.AddLine(wxString::Format("%.15g;%g;%g;%.15g;%g;%g",
					param[i], dparam[i], sparam[i],
					param[i+1], dparam[i+1], sparam[i+1]));
			}
			csvFile.Write();
			csvFile.Close();
		} else {
			wxMessageBox("Cannot create CSV file", "CpFit", wxICON_ERROR | wxOK);
		}
	}
}

/**
 * \brief Exports parameters, their confidence intervals and
 * some another data into HTML file (report)
 */
void GraphWindow::OnSaveAsHTML(wxCommandEvent &event)
{
	wxFileDialog openFileDialog(this, wxT("Save image"), "", "",
                       "html files (*.html)|*.html", wxFD_SAVE);
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		wxString fileName = openFileDialog.GetPath();
		wxTextFile htmlFile;
		if (htmlFile.Create(fileName)) {
			wxDateTime dtNow = wxDateTime(wxDateTime::GetTimeNow());
			// General header
			htmlFile.AddLine("<html><head><meta charset = \"utf-8\"/></head><body>");
			htmlFile.AddLine("<h1>Report</h1>");
			htmlFile.AddLine("<h2>General info</h2>");
			htmlFile.AddLine("<p>Date: " + dtNow.FormatDate() + " </p>");
			htmlFile.AddLine("<p>Time: " + dtNow.FormatTime() + " </p>");
			htmlFile.AddLine(wxString::Format("<p>Total number of points: %d", lsq.data.size()));
			// Baseline parameters table header
			htmlFile.AddLine("<h2>Optimized baseline parameters set</h2>");
			htmlFile.AddLine("<table>");
			htmlFile.AddLine("<style>");
			htmlFile.AddLine("table {border-collapse: collapse; border: 4px double black;}");
			htmlFile.AddLine("td {border: 1px solid black;}");
			htmlFile.AddLine("</style>");
			htmlFile.AddLine("<tr><td>&alpha;</td> <td>&Delta;(&alpha;)</td> <td><i>s</i>(&alpha;)</td>");
			htmlFile.AddLine("<td>&theta;&nbsp;/&nbsp;K</td> <td>&Delta;(&theta;)&nbsp;/&nbsp;K</td> <td><i>s</i>(&theta;)&nbsp;/&nbsp;K</td> </tr>");
			// Baseline parameters table content
			std::vector <double> param = lsqRes.GetParam();
			std::vector <double> sparam = lsqRes.GetParamStd();
			std::vector <double> dparam = lsqRes.GetParamConfInt();
			for (int i = 0; i < lsqRes.GetNumOfParams(); i += 2) {
				htmlFile.AddLine(wxString::Format("<tr><td>%.15g</td><td>%g</td><td>%g</td>",
					param[i], dparam[i], sparam[i]));
				htmlFile.AddLine(wxString::Format("    <td>%.15g</td><td>%g</td><td>%g</td></tr>",
					param[i+1], dparam[i+1], sparam[i+1]));
			}
			// Baseline parameters table ending
			htmlFile.AddLine("</table>");

			// Extra terms parameters report			
			if (lsqRes.extraTerms.size() > 0) {
				// Extra terms parameters table header
				htmlFile.AddLine("<h2>Optimized extra terms parameters set</h2>");
				// Extra terms parameters tables
				for (int i = 0; i < lsqRes.extraTerms.size(); i++) {
					ExtraTerm &term = lsqRes.extraTerms[i];
					htmlFile.AddLine(wxString::Format("<h3>Term %d; Type: %s</h3>\n", i + 1, term.GetTypeAsString()));
					// a) Tmin and Tmax (constants)
					htmlFile.AddLine(wxString::Format("<p><i>T</i><sub>min</sub> = %g &nbsp;K</p>\n", term.Tmin));
					htmlFile.AddLine(wxString::Format("<p><i>T</i><sub>min</sub> = %g &nbsp;K</p>\n", term.Tmax));
					// b) Constant parameters
					htmlFile.AddLine("<p><b>Constant parameters</b></p>\n");
					htmlFile.AddLine("<table>");
					htmlFile.AddLine("<style>");
					htmlFile.AddLine("table {border-collapse: collapse; border: 4px double black;}");
					htmlFile.AddLine("td {border: 1px solid black;}");
					htmlFile.AddLine("</style>");
					htmlFile.AddLine("<tr><td><i>i</i></td> <td>Name</td> <td><i>c<sub>i</sub></i></td></tr>\n");
					std::vector <double> fixparam = term.GetFixParam();
					std::vector <std::string> fixparam_names = term.GetFixParamNames();
					for (int j = 0; j < fixparam.size(); j++) {
						htmlFile.AddLine(wxString::Format(
							"<tr><td>%d</td> <td>%s</td> <td>%.15g</td></tr>\n",
							j + 1, fixparam_names[j], fixparam[j]
						));
					}
					htmlFile.AddLine("</table>");
					// c) LSQ parameters (optimized)
					htmlFile.AddLine("<p><b>LSQ parameters</b></p>\n");
					htmlFile.AddLine("<table>");
					htmlFile.AddLine("<style>");
					htmlFile.AddLine("table {border-collapse: collapse; border: 4px double black;}");
					htmlFile.AddLine("td {border: 1px solid black;}");
					htmlFile.AddLine("</style>");
					htmlFile.AddLine("<tr><td><i>i</i></td> <td>Name</td> <td><i>b<sub>i</sub></i></td> <td><i>s</i>(<i>b<sub>i</sub></i>)</td> <td>&Delta;(<i>b<sub>i</sub></i>)</td></tr>\n");
					std::vector <double> lsqparam = term.GetLSQParam();
					std::vector <double> lsqparam_std = term.GetLSQParamStd();
					std::vector <double> lsqparam_ci = term.GetLSQParamCi();
					std::vector <std::string> lsqparam_names = term.GetLSQParamNames();
					for (int j = 0; j < lsqparam.size(); j++) {
						htmlFile.AddLine(wxString::Format(
							"<tr><td>%d</td> <td>%s</td> <td>%.15g</td> <td>%.15g</td> <td>%.15g</td></tr>\n",
							j + 1, lsqparam_names[j], lsqparam[j], lsqparam_std[j], lsqparam_ci[j]
						));
					}
					htmlFile.AddLine("</table>\n");
				}
			}		
			// Deviations tables
			wxString fmtDevHdr = "<h2>%s</h2><table><style>"
				"table {border-collapse: collapse; border: 4px double black;}"
				"td {border: 1px solid black;}</style>";

			// a) heat capacities (Cp)
			//   header
			htmlFile.AddLine(wxString::Format(fmtDevHdr, "<i>C<sub>p</sub></i> deviations"));
			//   body
			LSQDevInfo devCp = lsqRes.GetYDeviations(DATATYPE_CP);
			htmlFile.AddLine(wxString::Format("<tr><td>Standard absolute deviation, J/(mol*K)</td>    <td>%g</td><td></td></tr>", devCp.stdAbsDev));
			htmlFile.AddLine(wxString::Format("<tr><td>Standard relative deviation, %%</td> <td>%g</td><td></td></tr>", devCp.stdRelDev * 100));
			htmlFile.AddLine(wxString::Format("<tr><td>Maximal absolute deviation, J/(mol*K)</td>     <td>%g</td><td>(at %g K)</td></tr>",
				devCp.maxAbsDev, devCp.maxAbsDevX));
			htmlFile.AddLine(wxString::Format("<tr><td>Maximal relative deviation, %%</td>  <td>%g</td><td>(at %g K)</td></tr>",
				devCp.maxRelDev * 100, devCp.maxRelDevX));
			//   ending
			htmlFile.AddLine("</table>");
			// b) enthalpies (H)
			//   header
			htmlFile.AddLine(wxString::Format(fmtDevHdr, "<h2><i>H</i>(<i>T<sub>2</sub></i>)-<i>H</i>(<i>T</i><sub>1</sub>) deviations</h2>"));
			//   body
			LSQDevInfo devH = lsqRes.GetYDeviations(DATATYPE_H);
			htmlFile.AddLine(wxString::Format("<tr><td>Standard absolute deviation, J/mol</td>    <td>%g</td><td></td></tr>", devH.stdAbsDev));
			htmlFile.AddLine(wxString::Format("<tr><td>Standard relative deviation, %%</td> <td>%g</td><td></td></tr>", devH.stdRelDev * 100));
			htmlFile.AddLine(wxString::Format("<tr><td>Maximal absolute deviation, J/mol</td>     <td>%g</td><td>(at %g K)</td></tr>",
				devH.maxAbsDev, devH.maxAbsDevX));
			htmlFile.AddLine(wxString::Format("<tr><td>Maximal relative deviation, %%</td>  <td>%g</td><td>(at %g K)</td></tr>",
				devH.maxRelDev * 100, devH.maxRelDevX));
			//   ending
			htmlFile.AddLine("</table>");
			// Final of the report
			htmlFile.AddLine("</body></html>");
			// Closing the file
			htmlFile.Write();
			htmlFile.Close();
		} else {
			wxMessageBox("Cannot create HTML file", "CpFit", wxICON_ERROR | wxOK);
		}
	}
}

/**
 * \brief Opens window for estimation of differences between obtained
 * approximation and Debye model.
 */
void GraphWindow::OnDebyeCorrection(wxCommandEvent &event)
{
	TDPEFunc *f = &lsqRes;
	DebyeCorrWindow *wnd = new DebyeCorrWindow(this, f);
	wnd->Show(true);
}


