/**
 * \file
 * \brief  Implementation of 2D canvas for drawing 2D plots (with
 * lines, dots, grid, axes labels and ticks etc.)
 *
 * \author Alexey L. Voskov (alvoskov@gmail.com)
 * \copyright (C) 2016-2017 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#ifndef __PLOTCANVAS_H
#define __PLOTCANVAS_H

class PlotCanvas;

/** Marker types (either none or its form) */
enum marker_t {
	MARKER_NONE, MARKER_CIRCLE, MARKER_SQUARE, MARKER_DIAMOND
};

/** Line type (either none or its form) */
enum line_t {
	LINE_NONE, LINE_SOLID, LINE_DOT, LINE_LONG_DASH, LINE_SHORT_DASH
};

/**
 * \brief Keeps properties of marker or line required for their rendering
 */
class PlotMarker
{
public:
	marker_t markerType; /*!< Marker type (none or its form) */
	line_t lineType; /*!< Line type (none or its form) */

	int radius; /*!< Marker radius, pixels */
	int lineWidth; /*!< Line width, pixels */
	wxColor lineColor; /*!< Line color */
	wxColor fillColor; /*!< Marker fill color */

	PlotMarker();
	PlotMarker(wxColor lineColor, wxColor fillColor,
		line_t lineType = LINE_SOLID, marker_t markerType = MARKER_NONE,
		int lineWidth = 2, int radius = 5);

	void Copy(PlotMarker &m);
	void Draw(wxDC &dc, wxPoint &xyScr, bool initPensAndBrushes = true);


	wxPen GetWxPen();
	wxBrush GetWxBrush();
};

/**
 * \brief  Contains data series: both data and data marker properties
 */
class DataSeries : public PlotMarker
{
public:
	DataSeries();
	DataSeries(const double *x, const double *y, int n);
	void Draw(PlotCanvas *pl, wxDC *dc);
	void DrawMarkers(PlotCanvas *pl, wxDC &dc);
	void DrawLines(PlotCanvas *pl, wxDC &dc);
	
	double GetMinX();
	double GetMaxX();
	double GetMinY();
	double GetMaxY();
	
	std::vector <wxRealPoint> xy; /*!< Data series points logical coordinates */
};


class PlotCanvasLegendEntry
{
public:
	PlotCanvasLegendEntry();
	PlotCanvasLegendEntry(wxString name, PlotMarker marker);
	wxString name;
	PlotMarker marker;
};

/**
 * \brief Implementation of canvas for plotting 2D scatter plots
 */
class PlotCanvas : public wxWindow
{
public:
	PlotCanvas(wxWindow *parent, wxWindowID id, const wxPoint &pos, const wxSize &size);
	wxPoint xyToScrXY(wxRealPoint xy);
	std::vector <wxPoint> xyToScrXY(std::vector <wxRealPoint> xy);
	
	void Plot(double *x, double *y, int n, PlotMarker marker);	
	void Plot(std::vector <double> x, std::vector <double> y, PlotMarker marker);
	bool SetAxesBox(double xmin, double xmax, double ymin, double ymax);
	void FitAxes();
	double GetAxesMinX();
	double GetAxesMaxX();
	void EnableGrid();
	void DisableGrid();
	void EnableLegend();
	void DisableLegend();

	void SetXLabel(wxString xLabel);
	void SetYLabel(wxString yLabel);
	wxString GetXLabel();
	wxString GetYLabel();

	int GetNumOfDataSeries();
	DataSeries GetDataSeries(int ind);

	bool SetLegend(std::vector <std::string> &seriesNames, std::vector <PlotMarker> &seriesMarkers, PlotMarker &defaultMarker);
	bool SetLegend(std::vector <wxString> &seriesNames, std::vector <PlotMarker> &seriesMarkers, PlotMarker &defaultMarker);
	
	bool IsPointInsideAxes(wxRealPoint xy);
	wxRealPoint FindLineAxesCrossing(wxRealPoint pointA, wxRealPoint pointB);
	
	
	void OnPaint( wxPaintEvent& event );
	void OnSize( wxSizeEvent& event );
	void OnMouseWheel( wxMouseEvent& event );
	void OnLeftMouseButton( wxMouseEvent& event );

	void Resize(wxSize size);
	void PaintToDC(wxDC &dc);
	bool SaveToBitmap(wxString fileName, wxBitmapType bmpType);
	bool SaveToSVG(wxString fileName);
	bool SaveToPS(wxString fileName);
	bool SaveToWMF(wxString fileName);
	bool SaveToM(wxString fileName);


	bool ToClipboard();
	
private:

	wxRect CalcLabelRect(wxDC &dc, wxString text);

	void DrawXTick(wxDC &dc, double x, wxString label);
	void DrawYTick(wxDC &dc, double y, wxString label);
	double DecimalRound(double val, int ndigits, int nmax, int isFloor);
	bool IsGoodAxisStep(double val);
	void DrawGrid(wxDC &dc);
	void DrawLegend(wxDC &dc);

	template <class S>
	bool SetLegendTemplate(std::vector <S> &seriesNames, std::vector <PlotMarker> &seriesMarkers, PlotMarker &defaultMarker);



	std::vector <DataSeries> dataSeries;
	wxColor bgColor;

	// Labels
	wxString xLabel;
	wxString yLabel;

	// Variables connected with LOGICAL coordinates
	wxRealPoint xy_min; // Crossing of the x and y axes (bottom left corner)
	wxRealPoint xy_max; // Upper right corner
	wxRealPoint xy_min_default; // Default value that is saved for OnMouseWheel
	wxRealPoint xy_max_default; // Default value that is saved for OnMouseWheel
	
	// Variables connected with PHYSICAL (SCREEN) coordinates
	wxPoint axisBoxPos; // Top-left corner coordinates of the axis box (pixels)
	wxSize  canvasSize; // Size of the entire canvas (DC)
	wxSize  axisBoxSize;  // Size of the axis box (pixels)
	wxPen   axisPen;
	wxPen   gridPen;
	wxFont  axisFont;
	wxFont  labelFont;
	wxFont  labelFontItalic;
	wxFont  legendFont;
	std::vector <double> xTicks;
	std::vector <double> yTicks;


	int marginTop, marginLeft;
	int marginBottom, marginRight;
	int tickLength;
	
	bool gridEnabled;
	bool legendEnabled;
	// Variables for implementation of legend
	std::vector<PlotCanvasLegendEntry> legendEntries;
	// Variables for implementation of drag-n-drop
	bool isDragging;
	wxPoint posDrag;
	// Event table
	wxDECLARE_EVENT_TABLE();
};

#endif
